from aisquared.base import BaseObject


class CustomRendering(BaseObject):

    def __init__(
            self,
            id: str,
            content_html: str,
            content_script: str,
            content_style: str,
            query_selector: str = None,

    ):
        super().__init__()
        self.id = id
        self.content_html = content_html
        self.content_script = content_script
        self.content_style = content_style
        self.query_selector = query_selector

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        if not isinstance(value, str):
            raise TypeError('id must be str')
        self._id = value

    @property
    def content_html(self):
        return self._content_html

    @content_html.setter
    def content_html(self, value):
        if not isinstance(value, str):
            raise TypeError('content_html must be str')
        self._content_html = value

    @property
    def content_script(self):
        return self._content_script

    @content_script.setter
    def content_script(self, value):
        if not isinstance(value, str):
            raise TypeError('content_script must be str')
        self._content_script = value

    @property
    def query_selector(self):
        return self._query_selector

    @query_selector.setter
    def query_selector(self, value):
        if not isinstance(value, str) and value is not None:
            raise TypeError('query_selector must be str or None')
        self._query_selector = value

    @property
    def content_style(self):
        return self._content_style

    @content_style.setter
    def content_style(self, value):
        if not isinstance(value, str):
            raise TypeError('content_style must be str')
        self._content_style = value

    def to_dict(self):
        return {
            'className': 'CustomRendering',
            'params': {
                'id': self.id,
                'querySelector': self.query_selector,
                'contentHtml': self.content_html,
                'contentScript': self.content_script,
                'contentStyle': self.content_style
            }
        }



from aisquared.base import BaseObject, DEFAULT_HTML_TAG_RENDERING_CSS, HTML_TAG_RENDERING_CSS_FILE
import platform
import json
import os


class HTMLTagRendering(BaseObject):
    """
    Rendering for HTML tags

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.rendering.HTMLTagRendering(
        'my HTML tag',
        'MyHTMLTagRenderingID',
        'MyContainerID',
        '<p>Example Text</p>',
        'extra_tag',
        'append',
        'name_key',
        'value_key',
        'name_value'
    )
    >>> my_obj.to_dict()
    {'className': 'HTMLTagRendering',
    'label': 'my HTML tag',
    'params': {'id': 'MyHTMLTagRenderingID',
    'containerId': 'MyContainerID',
    'htmlContent': '<p>Example Text</p>',
    'extraContentTag': 'extra_tag',
    'injectionAction': 'append',
    'predictionNameKey': 'name_key',
    'predictionValueKey': 'value_key',
    'predictionNameValue': 'name_value',
    'content': ''}}

    """

    def __init__(
        self,
        label: str,
        id: str,
        container_id: str,
        html_content: str,
        extra_content_tag: str,
        injection_action: str,
        prediction_name_key: str,
        prediction_value_key: str,
        prediction_name_value: str,
        content: str = '',
        css_params: dict = None
    ):
        """
        Parameters
        ----------
        label : str
            The label for the HTML tag
        id : str
            The ID for the HTML tag
        container_id : str
            The ID of the container
        html_content : str
            HTML to render
        extra_content_tag : str
            Any extra content tags to render
        injection_action : str
            The injection action
        prediction_name_key : str
            The key for the prediction name
        prediction_value_key : str
            The key for the prediction value
        prediction_name_values : str
            The value for the prediction name
        content : str (default '')
            The content
        css_params : dict or None (default None)
            Additional CSS parameters
        """
        super().__init__()
        self.label = label
        self.id = id
        self.container_id = container_id
        self.html_content = html_content
        self.extra_content_tag = extra_content_tag
        self.injection_action = injection_action
        self.prediction_name_key = prediction_name_key
        self.prediction_value_key = prediction_value_key
        self.prediction_name_value = prediction_name_value
        self.content = content

        if css_params is None:
            if os.path.exists(HTML_TAG_RENDERING_CSS_FILE):
                with open(HTML_TAG_RENDERING_CSS_FILE, 'r') as f:
                    self.css_params = json.load(f)
            else:
                self.css_params = DEFAULT_HTML_TAG_RENDERING_CSS
        else:
            if css_params.get('style'):
                self.css_params = css_params
            else:
                self.css_params = {'style': css_params}

    def to_dict(self) -> dict:
        """
        Return the configuration object as a dictionary
        """
        return {
            'className': 'HTMLTagRendering',
            'label': self.label,
            'params': {
                'id': self.id,
                'containerId': self.container_id,
                'htmlContent': self.html_content,
                'extraContentTag': self.extra_content_tag,
                'injectionAction': self.injection_action,
                'predictionNameKey': self.prediction_name_key,
                'predictionValueKey': self.prediction_value_key,
                'predictionNameValue': self.prediction_name_value,
                'content': self.content,
                'style': self.css_params['style']
            }
        }



from typing import Union
from aisquared.base import BaseObject


class Regression(BaseObject):
    """
    Postprocessing configuration object for Regression

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.postprocessing.Regression(
        10,
        100
    )
    >>> my_obj.to_dict()
    {'className': 'Regression', 'params': {'min': 10, 'max': 100, 'round': False}}

    """

    def __init__(
        self,
        min: Union[int, float] = None,
        max: Union[int, float] = None,
        round: bool = False
    ):
        """
        Parameters
        ----------
        min : None, int, or float (default None)
            The value to map an output of 0 to from the model
        max : None, int, or float (default None)
            The value to map an output of 1 to from the model
        round : bool (default False)
            Whether to round the output
        """
        super().__init__()
        self.min = min
        self.max = max
        self.round = round

    @property
    def min(self):
        return self._min

    @min.setter
    def min(self, value):
        if not isinstance(value, (float, int)) and value is not None:
            raise TypeError('min must be None, float, or int')
        self._min = value

    @property
    def max(self):
        return self._max

    @max.setter
    def max(self, value):
        if not isinstance(value, (float, int)) and value is not None:
            raise TypeError('max must be None, float, or int')
        self._max = value

    @property
    def round(self):
        return self._round

    @round.setter
    def round(self, value):
        if not isinstance(value, bool):
            raise TypeError('round must be Boolean valued')
        self._round = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'Regression',
            'params': {
                'min': self.min,
                'max': self.max,
                'round': self.round
            }
        }



from aisquared.base import BaseObject


class RegressionFeedback(BaseObject):
    """
    Feedback for regression

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.feedback.RegressionFeedback()
    >>> my_obj.to_dict()
    {'className': 'RegressionFeedback', 'params': {}}

    """

    def __init__(self):
        super().__init__()

    def to_dict(self) -> dict:
        """
        Return the object as a dictionary
        """
        return {
            'className': 'RegressionFeedback',
            'params': dict()
        }



from aisquared.base import BaseObject


class LocalModel(BaseObject):
    """
    Interaction with a model currently saved to the local
    file system

    Example usage:

    >>> import aisquared
    >>> analytic = aisquared.config.analytic.LocalModel(
        'model_path',
        'text'
    )
    >>> analytic.to_dict()
    {'className': 'LocalModel',
    'params': {'path': 'model_path',
    'inputType': 'text'}}

    """

    def __init__(
        self,
        path: str,
        input_type: str
    ):
        """
        Parameters
        ----------
        path : str or path-like or file-like
            The file path of the saved model
        input_type : str
            Input type to the model. Must be one of 'cv', 'text', or 'tabular'
        """
        super().__init__()
        self.path = path
        self.input_type = input_type

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        self._path = value

    @property
    def input_type(self):
        return self._input_type

    @input_type.setter
    def input_type(self, value):
        self._input_type = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'LocalModel',
            'params': {
                'path': self.path,
                'inputType': self.input_type
            }
        }



import pytest
import aisquared


def test_image_harvester_init():
    harvester = aisquared.config.harvesting.ImageHarvester()
    assert harvester.to_dict() == {
        'className': 'ImageHarvester',
        'params': {
            'how': 'all'
        }
    }


def test_image_harvester_dtype():
    with pytest.raises(ValueError):
        harvester = aisquared.config.harvesting.ImageHarvester('not_all')


def test_text_harvester():
    harvester = aisquared.config.harvesting.TextHarvester()
    assert harvester.to_dict() == {
        'className': 'TextHarvester',
        'params': {
            'how': 'all',
            'regex': None,
            'flags': 'gu',
            'bodyOnly': False,
            'limit': None
        }
    }
    harvester = aisquared.config.harvesting.TextHarvester(
        how='regex',
        regex='test',
        body_only=True
    )
    assert harvester.to_dict() == {
        'className': 'TextHarvester',
        'params': {
            'how': 'regex',
            'regex': 'test',
            'flags': 'gu',
            'bodyOnly': True,
            'limit': None
        }
    }


def test_text_harvester_regex():
    harvester = aisquared.config.harvesting.TextHarvester(
        how='regex',
        regex='\D(\d{5})\D'
    )
    assert harvester.to_dict() == {
        'className': 'TextHarvester',
        'params': {
            'how': 'regex',
            'regex': '\D(\d{5})\D',
            'flags': 'gu',
            'bodyOnly': False,
            'limit': None
        }
    }


def test_input_harvester():
    harvester = aisquared.config.harvesting.InputHarvester()
    assert harvester.to_dict() == {
        'className': 'InputHarvester',
        'params': {
            'inputType': 'text',
            'maxLength': None,
            'features': None
        }
    }
    harvester = aisquared.config.harvesting.InputHarvester(
        'tabular',
        features=[
            {
                'name': 'age',
                'dtype': 'numeric'
            }
        ]
    )
    assert harvester.to_dict() == {
        'className': 'InputHarvester',
        'params': {
            'inputType': 'tabular',
            'maxLength': None,
            'features': [
                {
                    'name': 'age',
                    'dtype': 'numeric'
                }
            ]
        }
    }


def test_input_harvester_fails():
    with pytest.raises(ValueError):
        aisquared.config.harvesting.InputHarvester('foo')

        aisquared.config.harvesting.InputHarvester(
            max_length=-1
        )

        aisquared.config.harvesting.InputHarvester(
            'tabular', features=['test'])


def test_keyword_harvester():
    harvester = aisquared.config.harvesting.TextHarvester(
        how='keywords',
        keywords=['foo', 'bar'],
        flags='gui'
    )
    assert harvester.to_dict() == {
        'className': 'TextHarvester',
        'params': {
            'bodyOnly': False,
            'how': 'regex',
            'regex': 'foo|bar',
            'flags': 'gui',
            'limit': None
        }
    }


def test_queryparameter_harvester():
    harvester = aisquared.config.harvesting.QueryParameterHarvester(
        'test',
        '*',
        'test'
    )
    assert harvester.to_dict() == {
        'className': 'QueryParameterHarvester',
        'params': {
            'queryKeys': ['test'],
            'urlLocations': ['*'],
            'attributes': ['test']
        }
    }


def test_chatbot_harvester():
    harvester = aisquared.config.harvesting.ChatbotHarvester(
        'ChatBot',
        max_length=256
    )
    assert harvester.to_dict() == {
        'className': 'ChatbotHarvester',
        'params': {
            'title': 'ChatBot',
            'maxLength': 256,
            'inputType': 'text',
            'features': None,
            'harvestHistory': False
        }
    }



from aisquared.base import BaseObject


class CustomObject(BaseObject):
    """
    Custom class that allows the user to define custom classes for configuration

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.base.CustomObject(
        'MyClass',
        key1 = 'foo',
        key2 = 'bar'
        )
    >>> my_obj.to_dict()
    {'className': 'MyClass', 'params': {'key1': 'foo', 'key2': 'bar'}}
    )
    """

    def __init__(self, class_name: str, top_level_kwargs: dict = None, **kwargs):
        """
        Parameters
        ----------
        class_name : str
            The name for the class
        top_level_kwargs : dict or None (default None)
            Any top-level keyword arguments to place above the `params` dictionary
        **kwargs : additional keyword arguments
            Other parameters to be used
        """
        super().__init__()
        self.class_name = class_name
        self.params = kwargs
        self.path = kwargs.get('path')
        self.top_level_kwargs = top_level_kwargs

    def to_dict(self) -> dict:
        """
        Get the object as a dictionary
        """
        config = {
            'className': self.class_name,
            'params': self.params
        }
        if isinstance(self.top_level_kwargs, dict):
            for k, v in self.top_level_kwargs.items():
                config[k] = v
        return config



from aisquared.base import BaseObject
import base64


class OnnxModel(BaseObject):
    """
    Run an ONNX model locally
    """

    def __init__(
            self,
            path: str,
            input_shape: list,
            output_key: str,
            return_key: str = None,
            input_type: str = 'text'
    ):
        super().__init__()
        self.path = path
        self.input_shape = input_shape
        self.output_key = output_key
        self.return_key = return_key
        self.input_type = input_type

        with open(self.path, 'rb') as f:
            data = f.read()
        self.onnx_data = base64.b64encode(data).decode('ascii')

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        if not isinstance(value, str):
            raise TypeError(f'path must be str, got {type(value)}')
        self._path = value

    @property
    def input_shape(self):
        return self._input_shape

    @input_shape.setter
    def input_shape(self, value):
        if not isinstance(value, list) and not all([isinstance(v, int) for v in value]):
            raise TypeError('input_shape must be a list of integers')
        self._input_shape = value

    @property
    def output_key(self):
        return self._output_key

    @output_key.setter
    def output_key(self, value):
        if not isinstance(value, str):
            raise TypeError('output_key must be str')
        self._output_key = value

    @property
    def return_key(self):
        return self._return_key

    @return_key.setter
    def return_key(self, value):
        if not isinstance(value, str) and value is not None:
            raise TypeError('return_key must be string or None')
        self._return_key = value

    @property
    def input_type(self):
        return self._input_type

    @input_type.setter
    def input_type(self, value):
        if value not in ['text', 'cv']:
            raise ValueError(
                f'input_type must be one of "text", "cv", got {value}')
        self._input_type = value

    @property
    def onnx_data(self):
        return self._onnx_data

    @onnx_data.setter
    def onnx_data(self, value):
        if not isinstance(value, str):
            raise TypeError('onnx_data must be a string')
        self._onnx_data = value

    def to_dict(self):
        return {
            'className': 'OnnxModel',
            'params': {
                'path': self.path,
                'inputShape': self.input_shape,
                'outputKey': self.output_key,
                'returnKey': self.return_key,
                'inputType': self.input_type,
                'onnxData': self.onnx_data
            }
        }



from aisquared.base import BaseObject, DEFAULT_CHART_RENDERING_CSS, CHART_RENDERING_CSS_FILE
import json
import os


class BarChartRendering(BaseObject):

    """
    Rendering class for rendering a Bar Chart

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.rendering.BarChartRendering(
        'my_label',
        'my_id',
        'my_bar_chart',
        'my_container_id',
        'name',
        'value',
        'name_value',
        True,
        'circle'
    )
    >>> my_obj.to_dict()
    {'className': 'BarChartRendering',
        'label': 'my_label',
        'params': {'id': 'my_id',
        'chartName': 'my_bar_chart',
        'containerId': 'my_container_id',
        'displayLegend': True,
        'legendIcon': 'circle',
        'width': 'auto',
        'height': 'auto',
        'xOffset': '0',
        'yOffset': '0',
        'datasource': [{'labels': None,
            'labelsKey': None,
            'consolidateRows': True,
            'predictionNameKey': 'name',
            'predictionValueKey': 'value',
            'predictionNameValue': 'name_value'}]}}
    """

    def __init__(
        self,
        label: str,
        id: str,
        chart_name: str,
        container_id: str,
        prediction_name_key: str,
        prediction_value_key: str,
        prediction_name_value: str,
        display_legend: bool,
        legend_icon: str,
        labels_key: str = None,
        width: str = 'auto',
        height: str = 'auto',
        xOffset: str = '0',
        yOffset: str = '0',
        labels: list = None,
        consolidate_rows: bool = True,
        css_params: dict = None
    ):
        """
        Parameters
        ----------
        label : str
            The label for the chart
        id : str
            The ID for the chart
        chart_name : str
            The name for the chart
        container_id : str
            The ID of the container to use
        prediction_name_key : str
            The key to use for the prediction name
        prediction_value_key : str
            The key to use for the prediction value
        prediction_name_value : str
            The value to use for the prediction name
        display_legend : bool
            Whether to display the chart legend
        legend_icon : str
            The legend icon to display
        labels_key : str
            The key to use for the labels
        width : str (default 'auto')
            The width of the chart
        height : str (default 'auto')
            The height of the chart
        xOffset : str (default '0')
            The offset on the x axis
        yOffset : str (default '0')
            The offset on the y axis
        labels : list of str or None (default None)
            Labels, if hard-coded
        consolidate_rows : bool (default True)
            Whether to consolidate rows in the data
        css_params: dict or None (default None)
            Additional CSS parameters

        """
        super().__init__()
        self.label = label
        self.id = id
        self.chart_name = chart_name
        self.labels = labels
        self.container_id = container_id
        self.prediction_name_key = prediction_name_key
        self.prediction_value_key = prediction_value_key
        self.prediction_name_value = prediction_name_value
        self.display_legend = display_legend
        self.legend_icon = legend_icon
        self.width = width
        self.height = height
        self.xOffset = xOffset
        self.yOffset = yOffset
        self.consolidate_rows = consolidate_rows
        self.labels_key = labels_key

        if css_params is None:
            if os.path.exists(CHART_RENDERING_CSS_FILE):
                with open(CHART_RENDERING_CSS_FILE, 'r') as f:
                    self.css_params = json.load(f)
            else:
                self.css_params = DEFAULT_CHART_RENDERING_CSS
        else:
            if css_params.get('style'):
                self.css_params = css_params
            else:
                self.css_params = {'style': css_params}

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'BarChartRendering',
            'label': self.label,
            'params': {
                'id': self.id,
                'chartName': self.chart_name,
                'containerId': self.container_id,
                'displayLegend': self.display_legend,
                'legendIcon': self.legend_icon,
                'width': self.width,
                'height': self.height,
                'xOffset': self.xOffset,
                'yOffset': self.yOffset,
                'datasource': [
                    {
                        'labels': self.labels,
                        'labelsKey': self.labels_key,
                        'consolidateRows': self.consolidate_rows,
                        'predictionNameKey': self.prediction_name_key,
                        'predictionValueKey': self.prediction_value_key,
                        'predictionNameValue': self.prediction_name_value,
                    }
                ],
                'style': self.css_params['style']
            }
        }



from .Steps import ZScore, MinMax, OneHot, DropColumn
from aisquared.base import BaseObject

ALLOWED_STEPS = (
    ZScore,
    MinMax,
    OneHot,
    DropColumn
)


class TabularPreprocesser(BaseObject):
    """
    Preprocesser object for tabular data

    Example usage:

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.tabular.TabularPreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.tabular.ZScore(
            [0, 1, 2],
            [0.2, 0.4, 0.6]
        )
    )
    """

    def __init__(
            self,
            steps: list = None
    ):
        """
        Parameters
        ----------
        steps : list or None (default None)
            List of preprocesser steps for tabular data
        """
        super().__init__()
        self.steps = None
        if steps is not None:
            for step in steps:
                self.add_step(step)

    def add_step(self, step):
        """
        Add a step to the preprocesser object
        """
        if not isinstance(step, ALLOWED_STEPS):
            raise TypeError(f'Each step must be one of {ALLOWED_STEPS}')
        if self.steps is None:
            self.steps = [step]
        else:
            self.steps = self.steps + [step]

    def to_dict(self):
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'TabularPreprocessor',
            'steps': [
                step.to_dict() for step in self.steps
            ]
        }



from aisquared.base import BaseObject


class ReverseMLWorkflow(BaseObject):
    """
    Creation of a ReverseML Workflow to interact with remote data sources
    """

    def __init__(
        self,
        label: str,
        connector_type: str,
        connector_action: str = 'read',
        input_type: str = 'text',
        filter_type: str = 'input',
        file_names: list = [],
        bucket: str = '',
        filter_by_columns: list = [],
        all: bool = False,
        arn: str = '',
        host: str = '',
        path: str = '',
        port: str = '',
        role: str = '',
        soql: str = '',
        query: str = '',
        token: str = '',
        column: str = '',
        db_name: str = '',
        db_user: str = '',
        period: int = None,
        schema: str = '',
        secret: str = '',
        account: str = '',
        data_map: list = [],
        db_table: str = '',
        client_id: str = '',
        file_name: str = '',
        password: str = '',
        schedule: str = '',
        sync_keys: dict = {'source': '', 'destination': ''},
        warehouse: str = '',
        data_source: str = '',
        cluster_name: str = '',
        client_secret: str = '',
        organization: str = '',
        authentication_type: str = ''
    ):
        self.label = label
        self.connector_type = connector_type
        self.connector_action = connector_action
        self.input_type = input_type
        self.filter_type = filter_type
        self.file_names = file_names
        self.bucket = bucket
        self.filter_by_columns = filter_by_columns
        self.all = all
        self.arn = arn
        self.host = host
        self.path = path
        self.port = port
        self.role = role
        self.soql = soql
        self.query = query
        self.token = token
        self.column = column
        self.db_name = db_name
        self.db_user = db_user
        self.period = period
        self.schema = schema
        self.secret = secret
        self.account = account
        self.data_map = data_map
        self.db_table = db_table
        self.client_id = client_id
        self.file_name = file_name
        self.password = password
        self.schedule = schedule
        self.sync_keys = sync_keys
        self.warehouse = warehouse
        self.data_source = data_source
        self.cluster_name = cluster_name
        self.client_secret = client_secret
        self.organization = organization
        self.authentication_type = authentication_type

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'ReverseMLWorkflow',
            'label': self.label,
            'params': {
                'connectorType': self.connector_type,
                'connectorAction': self.connector_action,
                'inputType': self.input_type,
                'filterType': self.filter_type,
                'fileNames': self.file_names,
                'bucket': self.bucket,
                'filterByColumns': self.filter_by_columns,
                'all': self.all,
                'arn': self.arn,
                'host': self.host,
                'path': self.path,
                'port': self.port,
                'role': self.role,
                'soql': self.soql,
                'query': self.query,
                'token': self.token,
                'column': self.column,
                'dbName': self.db_name,
                'dbUser': self.db_user,
                'period': self.period,
                'schema': self.schema,
                'secret': self.secret,
                'account': self.account,
                'dataMap': self.data_map,
                'dbTable': self.db_table,
                'clientId': self.client_id,
                'fileName': self.file_name,
                'password': self.password,
                'schedule': self.schedule,
                'syncKeys': self.sync_keys,
                'warehouse': self.warehouse,
                'dataSource': self.data_source,
                'clusterName': self.cluster_name,
                'clientSecret': self.client_secret,
                'organization': self.organization,
                'authenticationType': self.authentication_type
            }
        }



"""
The aisquared.logging subpackage contains utilities for performing experiments within aisquared.

This functionality is inhereted from MLFlow. Please see the MFLow documentatation at https://mlflow.org.
"""

try:
    from mlflow import *
except ImportError:
    pass



import pytest
import aisquared


def test_analytic_init():
    with pytest.raises(TypeError):
        aisquared.config.analytic.LocalAnalytic()
        aisquared.config.analytic.LocalModel()
        aisquared.config.analytic.DeployedAnalytic()
        aisquared.config.analytic.DeployedModel()
        aisquared.config.analytic.S3Connector()

        aisquared.config.analytic.LocalModel('test', 'foo')

    aisquared.config.analytic.LocalAnalytic(
        'test',
        'cv'
    )
    aisquared.config.analytic.LocalModel(
        'test',
        'text'
    )
    aisquared.config.analytic.DeployedAnalytic(
        'test',
        'POST',
        'text',
        body={'test': True}
    )
    aisquared.config.analytic.DeployedModel(
        'test',
        'cv'
    )
    aisquared.config.analytic.ReverseMLWorkflow(
        'bucket',
        ['filename1.csv'],
        'name',
        input_type='text',
        filter_type='static',
        filter_by_columns=[{
            'inputType': 'static',
            'columnName': 'status',
            'columnValue': 'pending'
        }],
        period=1
    )



from typing import Union
import numpy as np
import requests
import json


def get_remote_prediction(
    data: Union[dict, str, np.ndarray, list],
    host: str = '127.0.0.1',
    port: int = 2244
) -> list:
    """
    Send data to use for prediction

    Parameters
    ----------
    data : dict, str, np.ndarray, or list
        The data to be predicted on
    host : str (default '127.0.0.1')
        The host to use
    port : int (default '2244')
        The port to use

    Notes
    -----
    - If data is a dictionary, it is expected to already be
      correctly formatted
    - If data is a string, it is expected to already be
      correctly formatted

    Returns
    -------
    predictions : list
        The predictions from the deployed model
    """
    # Setup the url and headers
    url = f'http://{host}:{port}/predict'
    headers = {
        'Content-Type': 'application/json'
    }

    # Format the data
    if isinstance(data, dict):
        data = json.dumps(data)
    elif isinstance(data, str):
        data = data
    elif isinstance(data, np.ndarray):
        data = json.dumps(
            {
                'data': data.tolist()
            }
        )
    elif isinstance(data, list):
        data = json.dumps(
            {
                'data': data
            }
        )

    # Make the request
    with requests.session() as sess:
        resp = sess.post(
            url,
            data=data,
            headers=headers
        )

    if resp.status_code != 200:
        return resp
    else:
        return resp.json()['predictions']



from typing import Union
from aisquared.base import BaseObject


class ZScore(BaseObject):
    """
    Z-Score normalization preprocessing step

    Z-Score normalization takes each supplied column value, subtracts that column's provided mean, and divides
    by the provided standard deviation.

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.tabular.TabularPreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.tabular.ZScore(
            [0, 1, 2],
            [0.2, 0.4, 0.6]
        )
    )
    """

    def __init__(
            self,
            means: list,
            stds: list,
            columns: Union[int, list] = None
    ):
        """
        means : list
            List of integer or float values which are the means of the associated columns
        stds : list
            List of integer or float values which are the standard deviations of the associated columns
        columns : None or list (default None)
            If provided, is a list of column indexes to apply normalization to
        """
        super().__init__()
        self.means = means
        self.stds = stds
        self.columns = columns

        if len(self.means) != len(self.stds):
            raise ValueError('means and stds must have the same length')
        if self.columns is not None:
            if len(self.columns) != len(self.means):
                raise ValueError(
                    'Number of columns must match number of means and stds')

    @property
    def means(self):
        return self._means

    @means.setter
    def means(self, value):
        if not isinstance(value, list):
            raise TypeError('means must be a list')
        if not all([isinstance(val, (int, float)) for val in value]):
            raise TypeError('Each value in means must be int or float')
        self._means = value

    @property
    def stds(self):
        return self._stds

    @stds.setter
    def stds(self, value):
        if not isinstance(value, list):
            raise TypeError('stds must be a list')
        if not all([isinstance(val, (int, float)) for val in value]):
            raise TypeError('Each value in stds must be int or float')
        self._stds = value

    @property
    def columns(self):
        return self._columns

    @columns.setter
    def columns(self, value):
        if not isinstance(value, list) and value is not None:
            raise TypeError('If provided, columns must be list')
        if isinstance(value, list) and not all([isinstance(val, int) for val in value]):
            raise TypeError('Each value of columns must be an int')
        self._columns = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'ZScore',
            'params': {
                'means': self.means,
                'stds': self.stds,
                'columns': self.columns
            }
        }


class MinMax(BaseObject):
    """
    Min-Max Scaling preprocessing step

    Min-Max Scaling takes all associated columns and maps values relative to the minimum and maximum values
    of the training data.

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.tabular.TabularPreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.tabular.MinMax(
            [0, 1.1, 2],
            [0.2, 14, 18.3]
        )
    )
    """

    def __init__(
            self,
            mins: list,
            maxs: list,
            columns: list = None
    ):
        """
        Parameters
        ----------
        mins : list
            List of integers or floats associated with the minimum values of each column in the training data
        maxs : list
            List of integers or floats associated with the maximum values of each column in the training data
        columns : None or list (default None)
            If provided, a list of column indexes to apply scaling to
        """
        super().__init__()
        self.mins = mins
        self.maxs = maxs
        self.columns = columns
        if len(self.mins) != len(self.maxs):
            raise ValueError('Length of mins and maxs must equal')
        if self.columns is not None:
            if len(self.mins) != len(self.columns):
                raise ValueError(
                    'Number of mins and maxs must equal the number of columns')

    @property
    def mins(self):
        return self._mins

    @mins.setter
    def mins(self, value):
        if not isinstance(value, list):
            raise TypeError('mins must be a list')
        if not all([isinstance(val, (int, float)) for val in value]):
            raise TypeError('Each value in mins must be int or float')
        self._mins = value

    @property
    def maxs(self):
        return self._maxs

    @maxs.setter
    def maxs(self, value):
        if not isinstance(value, list):
            raise TypeError('maxs must be a list')
        if not all([isinstance(val, (int, float)) for val in value]):
            raise TypeError('Each value in maxs must be int or float')
        self._maxs = value

    @property
    def columns(self):
        return self._columns

    @columns.setter
    def columns(self, value):
        if not isinstance(value, list) and value is not None:
            raise TypeError('If passed, columns must be list')
        if value is not None:
            if not all([isinstance(val, int) for val in value]):
                raise TypeError(
                    'If passed, each value in columns must be an int')
        self._columns = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'MinMax',
            'params': {
                'mins': self.mins,
                'maxs': self.maxs,
                'columns': self.columns
            }
        }


class OneHot(BaseObject):
    """
    One Hot encoding preprocessing step

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.tabular.TabularPreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.tabular.OneHot(
            6,
            ['one', 'two', 'three']
        )
    )
    """

    def __init__(
            self,
            column: int,
            values: list
    ):
        """
        Parameters
        ----------
        column : int
            Integer index of the column to apply one hot encoding to
        values : list
            The values, in order, to create binary columns for. Note that if a default value is intended, that
            value should simply not be provided in this list
        """
        super().__init__()
        self.column = column
        self.values = values

    @property
    def column(self):
        return self._column

    @column.setter
    def column(self, value):
        if not isinstance(value, int):
            raise TypeError('column must be integer')
        self._column = value

    @property
    def values(self):
        return self._values

    @values.setter
    def values(self, value):
        if not isinstance(value, list):
            raise TypeError('values must be list')
        self._values = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'OneHot',
            'params': {
                'column': self.column,
                'values': self.values
            }
        }


class DropColumn(BaseObject):
    """
    Drop a column from tabular data

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.tabular.TabularPreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.tabular.DropColumn(
            3
        )
    )
    """

    def __init__(
            self,
            column: int
    ):
        """
        Parameters
        ----------
        column : int
            The column index to drop
        """
        super().__init__()
        self.column = column

    @property
    def column(self):
        return self._column

    @column.setter
    def column(self, value):
        if not isinstance(value, int):
            raise ValueError('column must be integer valued')
        self._column = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'DropColumn',
            'params': {
                'column': self.column
            }
        }



import pytest
import aisquared


def test_steps_init():
    aisquared.config.preprocessing.text.Tokenize()
    aisquared.config.preprocessing.text.RemoveCharacters()
    aisquared.config.preprocessing.text.ConvertToCase()
    aisquared.config.preprocessing.text.ConvertToVocabulary(
        {
            'this': 3,
            'is': 4,
            'a': 5,
            'test': 6
        }
    )
    aisquared.config.preprocessing.text.PadSequences()
    aisquared.config.preprocessing.text.Trim()

    aisquared.config.preprocessing.image.AddValue(10)
    aisquared.config.preprocessing.image.SubtractValue(10)
    aisquared.config.preprocessing.image.MultiplyValue(10)
    aisquared.config.preprocessing.image.DivideValue(10)
    aisquared.config.preprocessing.image.ConvertToColor('RGB')
    aisquared.config.preprocessing.image.Resize([10, 10])

    aisquared.config.preprocessing.tabular.ZScore([0, 0, 0], [1, 1, 1])
    aisquared.config.preprocessing.tabular.MinMax([0, 0, 0], [1, 1, 1])
    aisquared.config.preprocessing.tabular.OneHot(2, ['cat', 'dog', 'lizard'])
    aisquared.config.preprocessing.tabular.DropColumn(4)


def test_processers_init():
    tabular = aisquared.config.preprocessing.tabular.TabularPreprocesser()
    image = aisquared.config.preprocessing.image.ImagePreprocesser()
    text = aisquared.config.preprocessing.text.TextPreprocesser()

    tabular.add_step(
        aisquared.config.preprocessing.tabular.ZScore([0], [0])
    )
    image.add_step(
        aisquared.config.preprocessing.image.SubtractValue(10)
    )
    text.add_step(
        aisquared.config.preprocessing.text.RemoveCharacters()
    )

    with pytest.raises(TypeError):
        tabular.add_step(
            aisquared.config.preprocessing.text.RemoveCharacters()
        )
        image.add_step(
            aisquared.config.preprocessing.tabular.ZScore([0], [0])
        )
        text.add_step(
            aisquared.config.preprocessing.image.SubtractValue(10)
        )



from .AISquaredAPIException import AISquaredAPIException
from .additional_utils import _check_results_length
from aisquared.base import ENDPOINTS
import pandas as pd
import requests


def _create_user(
        url: str,
        headers: dict,
        user_name: str,
        given_name: str,
        family_name: str,
        email: str,
        role_id: str,
        active: bool,
        middle_name: str,
        company_id: str,
        password: str
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Create a user within the platform

    Parameters
    ----------
    user_name : str
        The display name of the user
    given_name : str
        The user's first name
    family_name : str
        The user's last name
    email : str
        The user's email
    role_id : str
        The ID of the role to be given to the user
    active : bool
        Whether the user is active
    middle_name : str or None
        The user's middle name
    company_id : str or None
        The user's company ID
    password : str or None
        The user's password
    """

    json_data = {
        'active': active,
        'userName': user_name,
        'givenName': given_name,
        'familyName': family_name,
        'email': email,
        'roleId': role_id
    }

    if middle_name:
        json_data['middleName'] = middle_name
    if company_id:
        json_data['coompanyId'] = company_id
    if password:
        json_data['password'] = password

    url = f'{url}/{ENDPOINTS["user"]}'

    with requests.Session() as sess:
        resp = sess.post(
            url,
            json=json_data,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())
    return resp.json()


def _update_user(
        url: str,
        headers: dict,
        user_id: str,
        user_name: str,
        given_name: str,
        family_name: str,
        email: str,
        role_id: str,
        active: bool,
        middle_name: str,
        company_id: str,
        password: str
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Update a user within the platform

    Parameters
    ----------
    user_name : str
        The display name of the user
    given_name : str
        The user's first name
    family_name : str
        The user's last name
    email : str
        The user's email
    role_id : str
        The ID of the role to be given to the user
    active : bool
        Whether the user is active
    middle_name : str or None
        The user's middle name
    company_id : str or None
        The user's company ID
    password : str or None
        The user's password
    """

    json_data = {
        'active': active,
        'userName': user_name,
        'givenName': given_name,
        'familyName': family_name,
        'email': email,
        'roleId': role_id
    }

    if middle_name:
        json_data['middleName'] = middle_name
    if company_id:
        json_data['coompanyId'] = company_id
    if password:
        json_data['password'] = password

    url = f'{url}/{ENDPOINTS["user"]}/{user_id}'

    with requests.Session() as sess:
        resp = sess.put(
            url,
            headers=headers,
            json=json_data
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())
    return resp.ok


def _delete_user(
        url: str,
        headers: dict,
        user_id: str
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Delete a user from the platform

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    user_id : string
        The id of the user you want to interact with
    """

    url = f'{url}/{ENDPOINTS["user"]}/{user_id}'

    with requests.Session() as sess:
        resp = sess.delete(
            url,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())
    return resp.ok


def _get_user(
        url: str,
        headers: dict,
        user_id: str
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Get information about a user from the platform

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    user_id : string
        The id of the user you want to interact with
    """

    url = f'{url}/{ENDPOINTS["user"]}/{user_id}'

    with requests.Session() as sess:
        resp = sess.get(
            url,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())
    return resp.json()


def _get_group(
        url=str,
        headers=dict,
        group_id=str
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Get information about a group from the platform

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    group_id : string
        The id of the group you want to interact with
    """

    url = f'{url}/{ENDPOINTS["group"]}/{group_id}'

    with requests.Session() as sess:
        resp = sess.get(
            url,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())
    return resp.json()


def _create_group(
        url: str,
        headers: dict,
        display_name: str,
        role_id: str
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Create a group within the platform

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    display_name : string
        The display name of the group
    group_id : string
        The role id for the group
    """

    url = f'{url}/{ENDPOINTS["group"]}'

    json_data = {
        'displayName': display_name,
        'roleId': role_id
    }

    with requests.Session() as sess:
        resp = sess.post(
            url,
            json=json_data,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())
    return resp.json()


def _delete_group(
        url: str,
        headers: dict,
        group_id: str
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Delete a group within the platform

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    group_id : string
        The id of the group
    """

    url = f'{url}/{ENDPOINTS["group"]}/{group_id}'

    with requests.Session() as sess:
        resp = sess.delete(
            url,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())
    return resp.ok


def _update_group(
        url: str,
        headers: dict,
        group_id: str,
        display_name: str,
        role_id: str
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Update a group within the platform

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    display_name : string
        The display name of the group
    role_id : string
        The role id for the group
    """

    url = f'{url}/{ENDPOINTS["group"]}/{group_id}'

    json_data = {
        'displayName': display_name,
        'roleId': role_id
    }

    with requests.Session() as sess:
        resp = sess.put(
            url,
            json=json_data,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())
    return resp.ok


def _users_to_group(
        url: str,
        headers: dict,
        group_id: str,
        user_ids: list,
        add: bool
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Add or remove users to / from a group

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    group_id : str
        The group to add the users to
    user_ids : list of str
        The IDs of the users to add
    add : bool
        Whether to add a user to a group (add = True) or remove a user from a group (add = False)
    """

    url = f'{url}/{ENDPOINTS["group_membership"]}'
    json_data = {
        'groupId': group_id,
        'userIds': user_ids
    }

    with requests.Session() as sess:
        if add:
            resp = sess.put(
                url,
                headers=headers,
                json=json_data
            )
        else:
            resp = sess.delete(
                url,
                headers=headers,
                json=json_data
            )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())
    return resp.ok


def _list_users(
        url: str,
        headers: dict,
        max_count: int,
        as_df: bool
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    List all users

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    max_count : int
        The maximum number of users to return
    as_df : bool
        Whether to return the data as a Pandas DataFrame
    """

    url = f'{url}/{ENDPOINTS["user_list"]}?count={max_count}&startIndex=1'

    with requests.Session() as sess:
        resp = sess.get(
            url,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())

    if as_df:
        df = pd.json_normalize(resp.json()['Resources'])
        _check_results_length(df)
        columns = ['id', 'userName', 'emails', 'active', 'groups',
                   'name.givenName', 'name.middleName', 'name.familyName']
        df = df[columns]

        # This is a change that should be fixed on the database side

        df['displayName'] = df['name.givenName'].astype(
            str) + ' ' + df['name.familyName']
        return df
    return resp.json()


def _list_groups(
        url: str,
        headers: dict,
        max_count: int,
        as_df: bool
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    List all groups

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    max_count : int
        The maximum number of users to return
    as_df : bool
        Whether to return the data as a Pandas DataFrame
    """

    url = f'{url}/{ENDPOINTS["group_list"]}?count={max_count}&startIndex=1'

    with requests.Session() as sess:
        resp = sess.get(
            url,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())

    if as_df:
        resp = resp.json()
        ids = [i['id'] for i in resp['Resources']]
        names = [i['displayName'] for i in resp['Resources']]
        members = []

        members = [[(u['value'], u['display'])
                    for u in i['members'] if i != []] for i in resp['Resources']]
        df = pd.DataFrame({'id': ids, 'name': names, 'members': members})
        _check_results_length(df)
        return df

    return resp.json()


def _list_group_users(
    url: str,
    headers: dict,
    as_df: bool,
    group_id: str
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    List all users in a group

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    max_count : int
        The maximum number of users to return
    as_df : bool
        Whether to return the data as a Pandas DataFrame
    """

    url = f'{url}/{ENDPOINTS["group_list"]}/{group_id}'

    with requests.Session() as sess:
        resp = sess.get(
            url,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())

    if as_df:
        resp = resp.json()
        ids = []
        names = []
        for d in resp['members']:
            ids.append(d['value'])
            names.append(d['display'])

        df = pd.DataFrame({'id': ids, 'username': names})
        _check_results_length(df)
        return df

    return resp.json()


def _list_roles(
        url: str,
        headers: dict,
        as_df: bool
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    List roles

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    as_df : bool
        Whether to return the data as a Pandas DataFrame
    """

    url = f'{url}/{ENDPOINTS["roles"]}'

    with requests.Session() as sess:
        resp = sess.get(
            url,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())

    if as_df:
        df = pd.DataFrame(resp.json()['content'])
        _check_results_length(df)
        return df
    return resp.json()['content']



from aisquared.base import DEFAULT_CONTAINER_RENDERING_CSS, DEFAULT_HTML_TAG_RENDERING_CSS, DEFAULT_TABLE_RENDERING_CSS, CONTAINER_RENDERING_CSS_FILE, HTML_TAG_RENDERING_CSS_FILE, TABLE_RENDERING_CSS_FILE
import json


def save_default_css():
    """
    Save default CSS so that default CSS can be edited and automatically utilized with changes

    Notes
    -----
    - Saves all CSS files to the `~/.aisquared/` directory
    """

    with open(CONTAINER_RENDERING_CSS_FILE, 'w') as f:
        json.dump(DEFAULT_CONTAINER_RENDERING_CSS, f)

    with open(HTML_TAG_RENDERING_CSS_FILE, 'w') as f:
        json.dump(DEFAULT_HTML_TAG_RENDERING_CSS, f)

    with open(TABLE_RENDERING_CSS_FILE, 'w') as f:
        json.dump(DEFAULT_TABLE_RENDERING_CSS, f)



from aisquared.base import BaseObject, POSITIONS, STATIC_POSITIONS, DEFAULT_CONTAINER_RENDERING_CSS, CONTAINER_RENDERING_CSS_FILE
import platform
import json
import os


class ContainerRendering(BaseObject):
    """
    Rendering for a container

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.rendering.ContainerRendering(
        'my container',
        'myContainerID',
        "[data-id='tabpanel-general']"
    )
    >>> my_obj.to_dict()
    {'className': 'ContainerRendering',
    'label': 'my container',
    'params': {'id': 'myContainerID',
    'width': 'auto',
    'height': 'auto',
    'display': 'flex',
    'xOffset': '0',
    'yOffset': '0',
    'position': 'absolute',
    'orientation': 'column',
    'querySelector': "[data-id='tabpanel-general']",
    'staticPosition': None}}

    """

    def __init__(
        self,
        label: str,
        id: str,
        query_selector: str,
        position: str = 'absolute',
        static_position: str = None,
        width: str = 'auto',
        height: str = 'auto',
        display: str = 'flex',
        xOffset: str = '0',
        yOffset: str = '0',
        orientation: str = 'column',
        css_params: dict = None
    ):
        """
        Parameters
        ----------
        label : str
            The label for the object
        id : str
            The id for the object
        query_selector : str
            Query selector for which panel to place the container in
        position : str (default 'absolute')
            The position to place the container in, either 'absolute' or 'static'
        static_position : str or None (default None)
            If `position` is 'static', must be provided, either 'prepend' or 'append'
        width : str (default 'auto')
            The width of the rendering
        height : str (default 'auto')
            The height of the rendering
        display : str (default 'flex')
            The type of display
        xOffset : str (default '0')
            The x offset of the rendering
        yOffset : str (default '0')
            The y offset of the rendering
        orientation : str (default 'column')
            The orientation of the rendering
        css_params : dict or None (default None)
            Additional CSS parameters
        """
        super().__init__()
        self.label = label
        self.id = id
        self.query_selector = query_selector
        self.position = position
        self.static_position = static_position
        self.width = width
        self.height = height
        self.display = display
        self.xOffset = xOffset
        self.yOffset = yOffset
        self.orientation = orientation

        if css_params is None:
            if os.path.exists(CONTAINER_RENDERING_CSS_FILE):
                with open(CONTAINER_RENDERING_CSS_FILE, 'r') as f:
                    self.css_params = json.load(f)
            else:
                self.css_params = DEFAULT_CONTAINER_RENDERING_CSS
        else:
            if css_params.get('style'):
                self.css_params = css_params
            else:
                self.css_params = {'style': css_params}

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        self._label = value

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def query_selector(self):
        return self._query_selector

    @query_selector.setter
    def query_selector(self, value):
        self._query_selector = value

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value):
        self._width = value

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, value):
        self._height = value

    @property
    def display(self):
        return self._display

    @display.setter
    def display(self, value):
        self._display = value

    @property
    def xOffset(self):
        return self._xOffset

    @xOffset.setter
    def xOffset(self, value):
        self._xOffset = value

    @property
    def yOffset(self):
        return self._yOffset

    @yOffset.setter
    def yOffset(self, value):
        self._yOffset = value

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, value):
        if value not in POSITIONS:
            raise ValueError(
                f'position must be one of {POSITIONS}, got {value}')
        self._position = value

    @property
    def orientation(self):
        return self._orientation

    @orientation.setter
    def orientation(self, value):
        self._orientation = value

    @property
    def static_position(self):
        return self._static_position

    @static_position.setter
    def static_position(self, value):
        if value is not None:
            if not isinstance(value, str):
                raise TypeError('If not None, static_position must be str')
            if value not in STATIC_POSITIONS:
                raise ValueError(
                    f'static_position must be one of {STATIC_POSITIONS}, got {value}')
        else:
            if self.position == 'static':
                raise ValueError(
                    'If position is "static", static_position must be provided')

        self._static_position = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'ContainerRendering',
            'label': self.label,
            'params': {
                'id': self.id,
                'width': self.width,
                'height': self.height,
                'display': self.display,
                'xOffset': self.xOffset,
                'yOffset': self.yOffset,
                'position': self.position,
                'orientation': self.orientation,
                'querySelector': self.query_selector,
                'position': self.position,
                'staticPosition': self.static_position,
                'style': self.css_params['style']
            }
        }



"""
The aisquared.base package contains both some basic objects that are used across the aisquared package backend and some objects which are designed to facilitate simple use cases of the technology.
"""

from .BaseObject import BaseObject
from .rendering import LOCATIONS, COLORS, BADGES, WORD_LISTS, QUALIFIERS, POSITIONS, STATIC_POSITIONS
from .stages import ALLOWED_STAGES
from .harvesting import ALLOWED_INPUT_TYPES, ALLOWED_HOWS
from .preprocessing import ALLOWED_PADS
from .css import DEFAULT_CONTAINER_RENDERING_CSS, DEFAULT_HTML_TAG_RENDERING_CSS, DEFAULT_TABLE_RENDERING_CSS, TABLE_RENDERING_CSS_FILE, HTML_TAG_RENDERING_CSS_FILE, CONTAINER_RENDERING_CSS_FILE, DEFAULT_CHART_RENDERING_CSS, CHART_RENDERING_CSS_FILE, DIRECTORY
from .platform import CLIENT_CONFIG_FILE
from .endpoints import ENDPOINTS

from aisquared.config.harvesting import ImageHarvester, TextHarvester, InputHarvester, QueryParameterHarvester, ChatbotHarvester
from aisquared.config.preprocessing.tabular import TabularPreprocesser
from aisquared.config.preprocessing.image import ImagePreprocesser
from aisquared.config.preprocessing.text import TextPreprocesser
from aisquared.config.analytic import DeployedAnalytic, DeployedModel, LocalModel, LocalAnalytic, ReverseMLWorkflow, OnnxModel
from aisquared.config.postprocessing import BinaryClassification, MulticlassClassification, ObjectDetection, Regression
from aisquared.config.rendering import ImageRendering, ObjectRendering, DocumentRendering, WordRendering, FilterRendering, ContainerRendering, HTMLTagRendering, DoughnutChartRendering, TableRendering, BarChartRendering, LineChartRendering, DashboardReplacementRendering, PieChartRendering, SOSRendering, TextRendering, CustomRendering, ChatRendering  # , DashboardRendering
from aisquared.config.feedback import SimpleFeedback, BinaryFeedback, MulticlassFeedback, RegressionFeedback, ModelFeedback, QualitativeFeedback
from aisquared.config.CustomObject import CustomObject

HARVESTING_CLASSES = (
    ImageHarvester,
    TextHarvester,
    InputHarvester,
    QueryParameterHarvester,
    ChatbotHarvester,
    CustomObject
)

PREPROCESSING_CLASSES = (
    TabularPreprocesser,
    ImagePreprocesser,
    TextPreprocesser,
    CustomObject
)

ANALYTIC_CLASSES = (
    DeployedAnalytic,
    DeployedModel,
    LocalModel,
    LocalAnalytic,
    ReverseMLWorkflow,
    OnnxModel,
    CustomObject
)

POSTPROCESSING_CLASSES = (
    BinaryClassification,
    MulticlassClassification,
    ObjectDetection,
    Regression,
    CustomObject
)

RENDERING_CLASSES = (
    ObjectRendering,
    ImageRendering,
    DocumentRendering,
    WordRendering,
    FilterRendering,
    ContainerRendering,
    HTMLTagRendering,
    DoughnutChartRendering,
    TableRendering,
    BarChartRendering,
    LineChartRendering,
    DashboardReplacementRendering,
    PieChartRendering,
    SOSRendering,
    TextRendering,
    CustomRendering,
    CustomObject,
    ChatRendering
    # DashboardRendering
)

FEEDBACK_CLASSES = (
    ModelFeedback,
    SimpleFeedback,
    BinaryFeedback,
    MulticlassFeedback,
    RegressionFeedback,
    QualitativeFeedback,
    CustomObject
)

LOCAL_CLASSES = (
    LocalModel,
    LocalAnalytic,
    OnnxModel,
    CustomObject
)



import json


class BaseObject:
    """
    Base class used for all other classes within the aisquared package. This class is not meant
    to be used by any end user of this package, but is rather used throughout this package as a
    parent class.
    """

    def __init__(self):
        pass

    def to_dict(self) -> dict:
        """
        Get the object as a dictionary
        """
        raise NotImplementedError

    def to_json(self) -> str:
        """
        Return the object as a json string
        """
        return json.dumps(self.to_dict())



from aisquared.base import BaseObject


class LocalAnalytic(BaseObject):
    """
    Interaction with an analytic (lookup table) saved to the
    local file system

    Example usage:

    >>> import aisquared
    >>> analytic = aisquared.config.analytic.LocalAnalytic(
        'analytic_path',
        'text'
    )
    >>> analytic.to_dict()
    {'className': 'LocalAnalytic',
    'params': {'path': 'analytic_path',
    'inputType': 'text',
    'all': False}}
    """

    def __init__(
        self,
        path: str,
        input_type: str,
        all: bool = False
    ):
        """
        Parameters
        ----------
        path : str or path-like or file-like
            The path to the analytic saved on disk
        input_type : str
            The input type to the analytic. Either one of 'cv', 'text', or 'tabular'
        all : bool (default False)
            Whether the entire analytic will be returned for every call
        """
        super().__init__()
        self.path = path
        self.input_type = input_type
        self.all = all

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        self._path = value

    @property
    def input_type(self):
        return self._input_type

    @input_type.setter
    def input_type(self, value):
        self._input_type = value

    @property
    def all(self):
        return self._all

    @all.setter
    def all(self, value):
        if not isinstance(value, bool):
            raise TypeError('all must be Boolean')
        self._all = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'LocalAnalytic',
            'params': {
                'path': self.path,
                'inputType': self.input_type,
                'all': self.all
            }
        }



"""
THIS MODULE IS IN DEVELOPMENT AND NOT STABLE. PLEASE USE WITH CAUTION AND DO NOT USE FOR ANY PRODUCTION WORKLOADS
"""

import random
import string

import warnings

from aisquared.base import BaseObject

from .BarChartRendering import BarChartRendering
from .ContainerRendering import ContainerRendering
from .DoughnutChartRendering import DoughnutChartRendering
from .HTMLTagRendering import HTMLTagRendering
from .LineChartRendering import LineChartRendering
from .PieChartRendering import PieChartRendering
from .TableRendering import TableRendering


class DashboardRendering(BaseObject):
    """
    THIS CLASS IS IN DEVELOPMENT AND IS NOT STABLE. PLEASE USE WITH CAUTION AND DO NOT USE FOR ANY PRODUCTION WORKLOADS
    """

    def __init__(self):
        super().__init__()
        self._steps = []
        warnings.warn(
            'THIS CLASS IS IN DEVELOPMENT AND IS NOT STABLE. PLEASE USE WITH CAUTION AND DO NOT USE FOR ANY PRODUCTION WORKLOADS',
            UserWarning
        )

    @property
    def steps(self):
        return self._steps

    def add_container(
        self,
        query_selector: str,
        width: str = 'auto',
        height: str = 'auto',
        display: str = 'flex',
        xOffset: str = '0',
        yOffset: str = '0',
        position: str = '',
        orientation: str = 'column',
        id: str = None,
        label: str = None
    ):
        if id is None:
            id = ''.join(random.choices(string.ascii_letters, k=10))
        if label is None:
            label = id

        self._steps.append(
            ContainerRendering(
                label=label,
                id=id,
                query_selector=query_selector,
                width=width,
                height=height,
                display=display,
                xOffset=xOffset,
                yOffset=yOffset,
                position=position,
                orientation=orientation
            )
        )

        return id

    def add_bar_chart(
        self,
        container_id: str,
        prediction_name_key: str,
        prediction_value_key: str,
        prediction_name_value: str,
        chart_colors: list,
        chart_labels: list,
        width: str = 'auto',
        height: str = 'auto',
        xOffset: str = '0',
        yOffset: str = '0',
        id: str = None,
        label: str = None,
        chart_name: str = None
    ):
        if id is None:
            id = ''.join(random.choices(string.ascii_letters, k=10))
        if label is None:
            label = id
        if chart_name is None:
            label = id

        self._steps.append(
            BarChartRendering(
                label=label,
                id=id,
                chart_name=chart_name,
                chart_colors=chart_colors,
                chart_labels=chart_labels,
                container_id=container_id,
                prediction_name_key=prediction_name_key,
                prediction_value_key=prediction_value_key,
                prediction_name_value=prediction_name_value,
                width=width,
                height=height,
                xOffset=xOffset,
                yOffset=yOffset
            )
        )

    def add_doughnut_chart(
        self,
        container_id: str,
        prediction_name_key: str,
        prediction_value_key: str,
        prediction_name_value: str,
        chart_colors: list,
        chart_labels: list,
        display_legend: bool = True,
        legend_icon: str = 'circle',
        width: str = 'auto',
        height: str = 'auto',
        xOffset: str = '0',
        yOffset: str = '0',
        id: str = None,
        label: str = None,
        chart_name: str = None
    ):
        if id is None:
            id = ''.join(random.choices(string.ascii_letters, k=10))
        if label is None:
            label = id
        if chart_name is None:
            chart_name = id

        self._steps.append(
            DoughnutChartRendering(
                label=label,
                id=id,
                chart_name=chart_name,
                chart_colors=chart_colors,
                chart_labels=chart_labels,
                container_id=container_id,
                prediction_name_key=prediction_name_key,
                prediction_value_key=prediction_value_key,
                prediction_name_value=prediction_name_value,
                display_legend=display_legend,
                legend_icon=legend_icon,
                width=width,
                height=height,
                xOffset=xOffset,
                yOffset=yOffset
            )
        )

    def add_html_tag(
        self,
        container_id: str,
        html_content: str,
        prediction_name_key: str = '',
        prediction_value_key: str = '',
        prediction_name_value: str = '',
        extra_content_tag: str = 'strong',
        injection_action: str = 'prepend',
        id: str = None,
        content: str = '',
        label: str = None
    ):
        if id is None:
            id = ''.join(random.choices(string.ascii_letters, k=10))
        if label is None:
            label = id

        self._steps.append(
            HTMLTagRendering(
                label=label,
                id=id,
                container_id=container_id,
                html_content=html_content,
                extra_content_tag=extra_content_tag,
                injection_action=injection_action,
                prediction_name_key=prediction_name_key,
                prediction_value_key=prediction_value_key,
                prediction_name_value=prediction_name_value,
                content=content
            )
        )

    def add_line_chart(
        self,
        container_id: str,
        prediction_name_key: str,
        prediction_value_key: str,
        prediction_name_value: str,
        chart_colors: list,
        chart_labels: list,
        width: str = 'auto',
        height: str = 'auto',
        xOffset: str = '0',
        yOffset: str = '0',
        id: str = None,
        label: str = None,
        chart_name: str = None
    ):
        if id is None:
            id = ''.join(random.choices(string.ascii_letters, k=10))
        if label is None:
            label = id
        if chart_name is None:
            chart_name = id

        self._steps.append(
            LineChartRendering(
                label=label,
                id=id,
                chart_name=chart_name,
                chart_colors=chart_colors,
                chart_labels=chart_labels,
                container_id=container_id,
                prediction_name_key=prediction_name_key,
                prediction_value_key=prediction_value_key,
                prediction_name_value=prediction_name_value,
                width=width,
                height=height,
                xOffset=xOffset,
                yOffset=yOffset
            )
        )

    def add_pie_chart(
        self,
        container_id: str,
        prediction_name_key: str,
        prediction_value_key: str,
        prediction_name_value: str,
        chart_colors: list,
        chart_labels: list,
        display_legend: bool = True,
        legend_icon: str = 'circle',
        width: str = 'auto',
        height: str = 'auto',
        xOffset: str = '0',
        yOffset: str = '0',
        id=None,
        label: str = None,
        chart_name: str = None
    ):
        if id is None:
            id = ''.join(random.choices(string.ascii_letters, k=10))
        if label is None:
            label = id
        if chart_name is None:
            chart_name = id

        self._steps.append(
            PieChartRendering(
                label=label,
                id=id,
                chart_name=chart_name,
                chart_colors=chart_colors,
                chart_labels=chart_labels,
                container_id=container_id,
                prediction_name_key=prediction_name_key,
                prediction_value_key=prediction_value_key,
                prediction_name_value=prediction_name_value,
                display_legend=display_legend,
                legend_icon=legend_icon,
                width=width,
                height=height,
                xOffset=xOffset,
                yOffset=yOffset
            )
        )

    def add_table(
        self,
        container_id: str,
        prediction_name_key: str,
        prediction_value_key: str,
        prediction_name_values: str,
        table_name: str = '',
        id: str = None,
        label: str = None
    ):
        if id is None:
            id = ''.join(random.choices(string.ascii_letters, k=10))
        if label is None:
            label = id

        self._steps.append(
            TableRendering(
                label=label,
                id=id,
                container_id=container_id,
                prediction_name_key=prediction_name_key,
                prediction_value_key=prediction_value_key,
                prediction_name_values=prediction_name_values,
                table_name=table_name
            )
        )

    def to_dict(self):
        return [
            step.to_dict() for step in self.steps
        ]



from aisquared.base import BaseObject


class DeployedAnalytic(BaseObject):
    """
    Interaction with a remote endpoint.

    Example usage:

    >>> import aisquared
    >>> analytic = aisquared.config.analytic.DeployedAnalytic(
        'model_url',
        'POST',
        'text',
        {
            'Content-Type' : 'application/json'
        },
        {
            'data_to_be_sent' : '{{input}}'
        },
        'string'
    )
    >>> analytic.to_dict()
    {'className': 'DeployedAnalytic',
    'params': {'url': 'model_url',
    'method': 'POST',
    'inputType': 'text',
    'headers': {'Content-Type': 'application/json'},
    'body': {'data_to_be_sent': '{{input}}'},
    'dataType': 'string'}}

    """

    def __init__(
        self,
        url: str,
        method: str,
        input_type: str,
        headers: dict = None,
        body: dict = None,
        data_type: str = 'string'
    ):
        """
        Parameters
        ----------
        url : str
            The base URL for the remote analytic
        method : str
            The method for hitting the API. Either 'POST' or 'GET'
        input_type : str
            The input types supplied to the analytic. Either one of 'cv' or 'text'
        headers : dict or None (default None)
            Header to use when calling the endpoint
        body : dict or None
            Prototype request body to be sent
        data_type : str
            The data type for the input variable to be converted to, one of 'string', 'array', 'object', or 'number'

        Notes
        -----
        - To input harvested data to the body, use the string "{{input}}"
        """
        super().__init__()
        self.url = url
        self.method = method
        self.input_type = input_type
        self.headers = headers
        self.body = body
        self.data_type = data_type

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, value):
        if not isinstance(value, str):
            raise TypeError(f'method must be str, got {type(value)}')
        if value not in ['GET', 'POST']:
            raise ValueError(
                f'method must be one of "GET", "POST", got {value}')
        self._method = value

    @property
    def input_type(self):
        return self._input_type

    @input_type.setter
    def input_type(self, value):
        self._input_type = value

    @property
    def headers(self):
        return self._headers

    @headers.setter
    def headers(self, value):
        self._headers = value

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, value):
        if not isinstance(value, dict):
            raise TypeError(f'body must be dict, got {type(value)}')
        self._body = value

    @property
    def data_type(self):
        return self._data_type

    @data_type.setter
    def data_type(self, value):
        if value not in ['string', 'array', 'object', 'number']:
            raise ValueError(
                f'data_type must be one of "string", "array", "object", or "number", got {value}')
        self._data_type = value

    def to_dict(self) -> dict:
        return {
            'className': 'DeployedAnalytic',
            'params': {
                'url': self.url,
                'method': self.method,
                'inputType': self.input_type,
                'headers': self.headers,
                'body': self.body,
                'dataType': self.data_type
            }
        }



"""
Some allowed configuration parameters - not designed to be directly called by the user
"""

ALLOWED_INPUT_TYPES = [
    'text',
    'image',
    'tabular'
]

ALLOWED_HOWS = [
    'all',
    'regex',
    'keywords'
]



"""
The aisquared.config.postprocessing subpackage contains objects for configuring how predictions are postprocessed.
"""

from .BinaryClassification import BinaryClassification
from .MulticlassClassification import MulticlassClassification
from .ObjectDetection import ObjectDetection
from .Regression import Regression



from aisquared.base import BaseObject
from .Steps import AddValue, SubtractValue, MultiplyValue, DivideValue, ConvertToColor, Resize

ALLOWED_STEPS = (
    AddValue,
    SubtractValue,
    MultiplyValue,
    DivideValue,
    ConvertToColor,
    Resize
)


class ImagePreprocesser(BaseObject):
    """
    Preprocesser object for image data

    Example usage:

    >>> import aisquared
    >>> preprocesser = aisquared.config.preprocessing.image.ImagePreprocesser()
    >>> preprocesser.add_step(
        aisquared.config.preprocessing.image.AddValue(255.0)
    )
    """

    def __init__(
            self,
            steps: list = None
    ):
        """
        Parameters
        ----------
        steps : list
            List of preprocessing steps for image data
        """
        super().__init__()
        self.steps = None
        if steps is not None:
            for step in steps:
                self.add_step(step)

    @property
    def step_dict(self):
        if self.steps is None:
            return self.steps
        else:
            return [
                step.to_dict() for step in self.steps
            ]

    def add_step(self, step):
        """
        Add a step to the preprocesser object
        """
        if not isinstance(step, ALLOWED_STEPS):
            raise TypeError(f'Each step must be one of {ALLOWED_STEPS}')
        if self.steps is None:
            self.steps = [step]
        else:
            self.steps = self.steps + [step]

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'ImagePreprocessor',
            'steps': self.step_dict
        }



from aisquared.base import BaseObject


class ChatRendering(BaseObject):
    """
    Rendering for a chatbot use case
    """

    def __init__(
            self,
            return_key,
            prediction_value_key=None,
            sender_name='You',
            responder_name='Chatbot'
    ):
        """
        Parameters
        ----------
        return_key : str
            The return key to retrieve from the returned text. If nested, use dot notation
            i.e. 'key1.key2.(...)'
        prediction_value_key : default None
            Deprecated, will be removed in a future version
        sender_name : str (default 'You')
            The name of the user persona in the displayed conversation
        responder_name : str (default 'Chatbot')
            The name of the non-user persona in the displayed conversation
        """
        super().__init__()
        self.return_key = return_key
        self.prediction_value_key = prediction_value_key
        self.sender_name = sender_name
        self.responder_name = responder_name

    @property
    def return_key(self):
        return self._return_key

    @return_key.setter
    def return_key(self, value):
        self._return_key = value

    @property
    def prediction_value_key(self):
        return self._prediction_value_key

    @prediction_value_key.setter
    def prediction_value_key(self, value):
        self._prediction_value_key = value

    @property
    def sender_name(self):
        return self._sender_name

    @sender_name.setter
    def sender_name(self, value):
        self._sender_name = value

    @property
    def responder_name(self):
        return self._responder_name

    @responder_name.setter
    def responder_name(self, value):
        self._responder_name = value

    def to_dict(self):
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'ChatRendering',
            'params': {
                'returnKey': self.return_key,
                'predictionValueKey': self.prediction_value_key,
                'senderName': self.sender_name,
                'responderName': self.responder_name
            }
        }



import pytest
import aisquared


def test_filter_rendering():
    with pytest.raises(TypeError):
        aisquared.config.rendering.FilterRendering()

    with pytest.raises(ValueError):
        aisquared.config.rendering.FilterRendering(
            source='test',
            key='test',
            qualifier='gt',
            value='test'
        )
        aisquared.config.rendering.FilterRendering(
            source='inputs',
            key='test',
            qualifier='test',
            value='test'
        )

    filter = aisquared.config.rendering.FilterRendering(
        source='inputs',
        key='test',
        qualifier='gt',
        value='test'
    )
    assert filter.to_dict() == {
        'className': 'FilterRendering',
        'params': {
            'source': 'inputs',
            'key': 'test',
            'qualifier': 'gt',
            'value': 'test'
        }
    }


def test_image_rendering():
    with pytest.raises(ValueError):
        aisquared.config.rendering.ImageRendering(
            placement='test'
        )

    image = aisquared.config.rendering.ImageRendering()

    assert image.to_dict() == {
        'className': 'ImageRendering',
        'params': {
            'color': 'blue',
            'thickness': '5',
            'placement': 'bottomleft',
            'includeProbability': False,
            'badgeColor': 'white',
            'fontColor': 'black',
            'fontSize': '5',
            'classes': None,
            'thresholdKey': None,
            'thresholdValue': None
        }
    }


def test_object_rendering():
    with pytest.raises(ValueError):
        aisquared.config.rendering.ObjectRendering(
            color='test'
        )
        aisquared.config.rendering.ObjectRendering(
            placement='test'
        )
        aisquared.config.rendering.ObjectRendering(
            badge_color='test'
        )
        aisquared.config.rendering.ObjectRendering(
            font_color='test'
        )

    object = aisquared.config.rendering.ObjectRendering()

    assert object.to_dict() == {
        'className': 'ObjectRendering',
        'params': {
            'color': 'blue',
            'thickness': '5',
            'placement': 'bottomleft',
            'includeProbability': False,
            'badgeColor': 'white',
            'fontColor': 'black',
            'fontSize': '5'
        }
    }


def test_word_rendering():
    with pytest.raises(ValueError):
        aisquared.config.rendering.WordRendering(
            word_list='test'
        )
        aisquared.config.rendering.WordRendering(
            badge_shape='test'
        )

    word = aisquared.config.rendering.WordRendering()
    assert word.to_dict() == {
        'className': 'WordRendering',
        'params': {
            'wordList': 'input',
            'resultKey': None,
            'contentKey': None,
            'badgeShape': 'star',
            'badgeColor': 'blue',
            'classes': None,
            'thresholdKey': None,
            'thresholdValue': None,
            'position': 'after'
        }
    }
    word = aisquared.config.rendering.WordRendering(badge_shape='underline')
    assert word.to_dict() == {
        'className': 'WordRendering',
        'params': {
            'wordList': 'input',
            'resultKey': None,
            'contentKey': None,
            'badgeShape': 'underline',
            'badgeColor': 'blue',
            'classes': None,
            'thresholdKey': None,
            'thresholdValue': None,
            'position': 'after'
        }
    }


def test_document_rendering():
    document = aisquared.config.rendering.DocumentRendering()
    assert document.to_dict() == {
        'className': 'DocumentRendering',
        'params': {
            'predictionKey': 'className',
            'words': None,
            'documents': None,
            'includeProbability': False,
            'probabilityKey': 'probability',
            'underlineColor': 'blue',
            'classes': None,
            'thresholdKey': None,
            'thresholdValue': None
        }
    }


def test_bar_chart_rendering():
    renderer = aisquared.config.rendering.BarChartRendering(
        label='test',
        id='test',
        chart_name='test',
        container_id='test',
        prediction_name_key='',
        prediction_value_key='',
        prediction_name_value='',
        display_legend=True,
        legend_icon='circle',
        labels_key='labels'
    )
    assert renderer.to_dict() == {'className': 'BarChartRendering',
                                  'label': 'test',
                                  'params': {'id': 'test',
                                             'chartName': 'test',
                                             'containerId': 'test',
                                             'displayLegend': True,
                                             'legendIcon': 'circle',
                                             'width': 'auto',
                                             'height': 'auto',
                                             'xOffset': '0',
                                             'yOffset': '0',
                                             'datasource': [{'labels': None,
                                                             'labelsKey': 'labels',
                                                             'consolidateRows': True,
                                                             'predictionNameKey': '',
                                                             'predictionValueKey': '',
                                                             'predictionNameValue': ''}],
                                             'style': {'container': {'color': 'black',
                                                                     'border': '1px solid lightgray',
                                                                     'margin': '0',
                                                                     'padding': '24px',
                                                                     'fontSize': '16px',
                                                                     'background': 'white',
                                                                     'borderRadius': '10px'}}}}


def test_line_chart_rendering():
    renderer = aisquared.config.rendering.LineChartRendering(
        label='test',
        id='test',
        chart_name='test',
        container_id='test',
        prediction_name_key='',
        prediction_value_key='',
        prediction_name_value='',
        display_legend=True,
        legend_icon='circle',
        labels_key='labels'
    )
    assert renderer.to_dict() == {'className': 'LineChartRendering',
                                  'label': 'test',
                                  'params': {'id': 'test',
                                             'chartName': 'test',
                                             'containerId': 'test',
                                             'displayLegend': True,
                                             'legendIcon': 'circle',
                                             'width': 'auto',
                                             'height': 'auto',
                                             'xOffset': '0',
                                             'yOffset': '0',
                                             'datasource': [{'labels': None,
                                                             'labelsKey': 'labels',
                                                             'consolidateRows': True,
                                                             'predictionNameKey': '',
                                                             'predictionValueKey': '',
                                                             'predictionNameValue': ''}],
                                             'style': {'container': {'color': 'black',
                                                                     'border': '1px solid lightgray',
                                                                     'margin': '0',
                                                                     'padding': '24px',
                                                                     'fontSize': '16px',
                                                                     'background': 'white',
                                                                     'borderRadius': '10px'}}}}


def test_doughnut_chart_rendering():
    renderer = aisquared.config.rendering.DoughnutChartRendering(
        label='test',
        id='test',
        chart_name='test',
        container_id='test',
        prediction_name_key='',
        prediction_value_key='',
        prediction_name_value='',
        display_legend=True,
        legend_icon='circle',
        labels_key='labels'
    )
    assert renderer.to_dict() == {'className': 'DoughnutChartRendering',
                                  'label': 'test',
                                  'params': {'id': 'test',
                                             'chartName': 'test',
                                             'containerId': 'test',
                                             'displayLegend': True,
                                             'legendIcon': 'circle',
                                             'width': 'auto',
                                             'height': 'auto',
                                             'xOffset': '0',
                                             'yOffset': '0',
                                             'datasource': [{'labels': None,
                                                             'labelsKey': 'labels',
                                                             'consolidateRows': True,
                                                             'predictionNameKey': '',
                                                             'predictionValueKey': '',
                                                             'predictionNameValue': ''}],
                                             'style': {'container': {'color': 'black',
                                                                     'border': '1px solid lightgray',
                                                                     'margin': '0',
                                                                     'padding': '24px',
                                                                     'fontSize': '16px',
                                                                     'background': 'white',
                                                                     'borderRadius': '10px'}}}}


def test_pie_chart_rendering():
    renderer = aisquared.config.rendering.PieChartRendering(
        label='test',
        id='test',
        chart_name='test',
        container_id='test',
        prediction_name_key='',
        prediction_value_key='',
        prediction_name_value='',
        display_legend=True,
        legend_icon='circle',
        labels_key='labels'
    )
    assert renderer.to_dict() == {'className': 'PieChartRendering',
                                  'label': 'test',
                                  'params': {'id': 'test',
                                             'chartName': 'test',
                                             'containerId': 'test',
                                             'displayLegend': True,
                                             'legendIcon': 'circle',
                                             'width': 'auto',
                                             'height': 'auto',
                                             'xOffset': '0',
                                             'yOffset': '0',
                                             'datasource': [{'labels': None,
                                                             'labelsKey': 'labels',
                                                             'consolidateRows': True,
                                                             'predictionNameKey': '',
                                                             'predictionValueKey': '',
                                                             'predictionNameValue': ''}],
                                             'style': {'container': {'color': 'black',
                                                                     'border': '1px solid lightgray',
                                                                     'margin': '0',
                                                                     'padding': '24px',
                                                                     'fontSize': '16px',
                                                                     'background': 'white',
                                                                     'borderRadius': '10px'}}}}


def test_container_rendering():
    renderer = aisquared.config.rendering.ContainerRendering(
        label='test',
        id='test',
        query_selector='test'
    )
    assert renderer.to_dict() == {'className': 'ContainerRendering',
                                  'label': 'test',
                                  'params': {'id': 'test',
                                             'width': 'auto',
                                             'height': 'auto',
                                             'display': 'flex',
                                             'xOffset': '0',
                                             'yOffset': '0',
                                             'position': 'absolute',
                                             'orientation': 'column',
                                             'querySelector': 'test',
                                             'staticPosition': None,
                                             'style': {'container': {'color': 'black',
                                                                     'margin': '0',
                                                                     'padding': '24px',
                                                                     'fontSize': '16px',
                                                                     'background': 'white',
                                                                     'borderRadius': '10px'}}}}


def test_dashboard_replacement_rendering():
    renderer = aisquared.config.rendering.DashboardReplacementRendering(
        'test'
    )
    assert renderer.to_dict() == {
        'className': 'DashboardReplacementRendering',
        'label': '',
        'params': {
            'anchorSelector': 'test',
            'whereReplace': ''
        }
    }


def test_html_tag_rendering():
    renderer = aisquared.config.rendering.HTMLTagRendering(
        label='test',
        id='test',
        container_id='test',
        html_content='test',
        extra_content_tag='test',
        injection_action='test',
        prediction_name_key='test',
        prediction_value_key='test',
        prediction_name_value='test',
    )

    assert renderer.to_dict() == {'className': 'HTMLTagRendering',
                                  'label': 'test',
                                  'params': {'id': 'test',
                                             'containerId': 'test',
                                             'htmlContent': 'test',
                                             'extraContentTag': 'test',
                                             'injectionAction': 'test',
                                             'predictionNameKey': 'test',
                                             'predictionValueKey': 'test',
                                             'predictionNameValue': 'test',
                                             'content': '',
                                             'style': {'content': {'border': 'none',
                                                                   'margin': '0',
                                                                   'padding': '8px',
                                                                   'background': 'transparent',
                                                                   'borderRadius': '0',
                                                                   'color': '#000000',
                                                                   'fontSize': '14px',
                                                                   'textAlign': 'left',
                                                                   'textDecoration': 'none',
                                                                   'textTransform': 'none',
                                                                   'fontWeight': 'bold',
                                                                   'width': 'auto',
                                                                   'height': 'auto'},
                                                       'extraContent': {'border': 'none',
                                                                        'margin': '0',
                                                                        'padding': '8px',
                                                                        'background': 'transparent',
                                                                        'borderRadius': '0',
                                                                        'color': '#000000',
                                                                        'fontSize': '14px',
                                                                        'textAlign': 'left',
                                                                        'textDecoration': 'none',
                                                                        'textTransform': 'none',
                                                                        'fontWeight': 'bold',
                                                                        'width': 'auto',
                                                                        'height': 'auto'}}}}


def test_sos_rendering():
    renderer = aisquared.config.rendering.SOSRendering(True)
    assert renderer.to_dict() == {
        'className': 'SOSRendering',
        'label': '',
        'params': {
            'canToggle': True
        }
    }


def test_table_rendering():
    renderer = aisquared.config.rendering.TableRendering(
        label='test',
        id='test',
        container_id='test',
        prediction_name_key='test',
        prediction_value_key='test',
        prediction_name_values=['name1', 'name2']
    )
    assert renderer.to_dict() == {'className': 'TableRendering',
                                  'label': 'test',
                                  'params': {'id': 'test',
                                             'containerId': 'test',
                                             'predictionNameKey': 'test',
                                             'predictionValueKey': 'test',
                                             'predictionNameValues': ['name1', 'name2'],
                                             'tableName': '',
                                             'style': {'container': {'border': '1px solid lightgray',
                                                                     'margin': '8px 0',
                                                                     'padding': '10px',
                                                                     'background': '#f3f3f3',
                                                                     'borderRadius': '10px',
                                                                     'color': '#000000',
                                                                     'fontSize': '14px',
                                                                     'textAlign': 'left',
                                                                     'fontFamily': 'Raleway, sans-serif'},
                                                       'title': {'fontWeight': 'bold', 'marginBottom': '4px'},
                                                       'table': {'border': '1px solid lightgray',
                                                                 'textAlign': 'center',
                                                                 'textTransform': 'capitalize',
                                                                 'width': '100%',
                                                                 'fontFamily': 'Raleway, sans-serif'},
                                                       'tableHeader': {'border': '1px solid lightgray',
                                                                       'padding': '4px',
                                                                       'fontWeight': 'bold',
                                                                       'textTransform': 'capitalize'},
                                                       'tableBody': {},
                                                       'tableRows': {'border': '1px solid lightgray'},
                                                       'rowCells': {'border': '1px solid lightgray',
                                                                    'fontWeight': 'normal',
                                                                    'padding': '8px'},
                                                       'headerCells': {'border': '1px solid lightgray',
                                                                       'padding': '8px',
                                                                       'fontWeight': 'bolder',
                                                                       'backgroundColor': 'white'}}}}


def test_text_renderer():
    renderer = aisquared.config.rendering.TextRendering()
    assert renderer.to_dict() == {
        'className': 'TextRendering',
        'params': {
            'predictionValueKey': None
        }
    }


def test_custom_renderer():
    renderer = aisquared.config.rendering.CustomRendering(
        'test',
        'test',
        'test',
        'test'
    )
    assert renderer.to_dict() == {
        'className': 'CustomRendering',
        'params': {
            'id': 'test',
            'querySelector': None,
            'contentHtml': 'test',
            'contentScript': 'test',
            'contentStyle': 'test'
        }
    }



from aisquared.base import BaseObject, ALLOWED_INPUT_TYPES


class InputHarvester(BaseObject):
    """
    Object to harvest user-input text

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.harvesting.InputHarvester()
    >>> my_obj.to_dict()
    {'className': 'InputHarvester',
    'params': {'inputType': 'text', 'maxLength': None, 'features': None}}

    """

    def __init__(
            self,
            input_type: str = 'text',
            max_length: int = None,
            features: list = None
    ):
        """
        Parameters
        ----------
        input_type : str (default 'text')
            What kind of input to harvest
            NOTE: Supports 'text', 'image', or 'tabular'
        max_length : int or None (default None)
            The maximum length of harvested text
        features : list or None (default None)
            Features to be harvested, if using 'tabular' harvesting.
            Otherwise ignored
        """
        super().__init__()
        self.input_type = input_type
        self.max_length = max_length
        self.features = features

    # input_type
    @property
    def input_type(self):
        return self._input_type

    @input_type.setter
    def input_type(self, value):
        if value not in ALLOWED_INPUT_TYPES:
            raise ValueError(
                "InputHarvester supports 'text', 'image', or 'tabular' input")
        self._input_type = value

    # max_length
    @property
    def max_length(self):
        return self._max_length

    @max_length.setter
    def max_length(self, value):
        if not isinstance(value, int) and value is not None:
            raise TypeError('max_length must be int or None')
        if isinstance(value, int) and value <= 0:
            raise ValueError('max_length must be positive')
        self._max_length = value

    # features
    @property
    def features(self):
        return self._features

    @features.setter
    def features(self, value):

        # Check that the value is not None
        if value is not None:

            # Check that the value is a list if not None
            if not isinstance(value, list):
                raise TypeError('features must be list or None')

            # Validate each instance in the list is a dictionary
            if not all([isinstance(val, dict) for val in value]):
                raise ValueError(
                    'Each value within `features` must be a dictionary')

            # Validate each of the names and dtypes for all features
            for val in value:
                if not val.get('name'):
                    raise ValueError('Each feature must have a `name` key')
                if val.get('dtype') not in ['string', 'numeric']:
                    raise ValueError(
                        'Each feature must have `dype` set to either `string` or `numeric`')

        self._features = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'InputHarvester',
            'params': {
                'inputType': self.input_type,
                'maxLength': self.max_length,
                'features': self.features
            }
        }



from typing import Union
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin, TransformerMixin
from sklearn.metrics import classification_report, confusion_matrix, mean_squared_error
import tensorflow as tf
import numpy as np
import beyondml


def _print_report(true_data, orig_preds, mimic_preds, problem_type):
    print('ORIGINAL PERFORMANCE:')
    if problem_type == 'classification':
        print(confusion_matrix(true_data, orig_preds))
        print(classification_report(true_data, orig_preds))
    else:
        print(mean_squared_error(true_data, orig_preds, squared=False))
    print('\n\n')

    print('MIMIC PERFORMANCE:')
    print('\n')
    print('Relative to First Model:')
    if problem_type == 'classification':
        if mimic_preds.shape[1] == 1:
            print(confusion_matrix(orig_preds, (mimic_preds >= 0.5).astype(int)))
            print(classification_report(orig_preds,
                  (mimic_preds >= 0.5).astype(int)))
        else:
            print(confusion_matrix(orig_preds, mimic_preds.argmax(axis=1)))
            print(classification_report(orig_preds, mimic_preds.argmax(axis=1)))
    else:
        print(mean_squared_error(orig_preds, mimic_preds, squared=False))
        print(f'Standard deviation: {np.std(mimic_preds - orig_preds)}')
    print('\n')

    print('Relative to Original:')
    if problem_type == 'classification':
        if mimic_preds.shape[1] == 1:
            print(confusion_matrix(true_data, (mimic_preds >= 0.5).astype(int)))
            print(classification_report(
                true_data, (mimic_preds >= 0.5).astype(int)))
        else:
            print(confusion_matrix(true_data, mimic_preds.argmax(axis=1)))
            print(classification_report(true_data, mimic_preds.argmax(axis=1)))
    else:
        print(mean_squared_error(true_data, mimic_preds, squared=False))
        print(f'Standard deviation: {np.std(true_data - orig_preds)}')


def _get_cv_model(size, input_shape, num_outputs, output_activation):
    if size == 'small':
        num_blocks = 2
        num_hidden = 1
        hidden_size = 64
    elif size == 'medium':
        num_blocks = 4
        num_hidden = 2
        hidden_size = 128
    elif size == 'large':
        num_blocks = 6
        num_hidden = 4
        hidden_size = 256
    else:
        raise ValueError(
            f'size must be one of "small", "medium", "large", got {size}')

    input_layer = tf.keras.layers.Input(input_shape)
    x = tf.keras.layers.Conv2D(
        8, 3, padding='same', activation='relu')(input_layer)
    x = tf.keras.layers.Conv2D(8, 3, padding='same', activation='relu')(x)
    x = tf.keras.layers.MaxPool2D(strides=1)(x)

    for block_num in range(num_blocks - 1):
        x = tf.keras.layers.Conv2D(
            8 * (2 ** (block_num + 1)), 3, padding='same', activation='relu')(x)
        x = tf.keras.layers.Conv2D(
            8 * (2 ** (block_num + 1)), 3, padding='same', activation='relu')(x)
        x = tf.keras.layers.MaxPool2D(strides=1)(x)

    x = tf.keras.layers.Flatten()(x)
    for _ in range(num_hidden):
        x = tf.keras.layers.Dense(hidden_size, activation='relu')(x)

    output_layer = tf.keras.layers.Dense(
        num_outputs, activation=output_activation)(x)

    return tf.keras.models.Model(input_layer, output_layer)


def _get_embedding_model(size, vocab_size, input_shape, num_outputs, output_activation):

    if vocab_size is None:
        raise ValueError(
            'If NLP embedding model specified, must also specify vocab_size')

    if size == 'small':
        embedding_dim = 4
        num_hidden = 4
        hidden_size = 64
    elif size == 'medium':
        embedding_dim = 8
        num_hidden = 6
        hidden_size = 128
    elif size == 'large':
        embedding_dim = 16
        num_hidden = 8
        hidden_size = 256
    else:
        raise ValueError(
            f'size must be one of "small", "medium", "large", got {size}')

    input_layer = tf.keras.layers.Input(input_shape)
    x = tf.keras.layers.Embedding(vocab_size, embedding_dim)(input_layer)
    x = tf.keras.layers.Flatten()(x)
    for _ in range(num_hidden):
        x = tf.keras.layers.Dense(hidden_size, activation='relu')(x)
    output_layer = tf.keras.layers.Dense(
        num_outputs, activation=output_activation)(x)

    return tf.keras.models.Model(input_layer, output_layer)


def _get_fc_model(size, input_shape, num_outputs, output_activation):
    if size == 'small':
        num_hidden = 4
        hidden_size = 64
    elif size == 'medium':
        num_hidden = 6
        hidden_size = 128
    elif size == 'large':
        num_hidden = 8
        hidden_size = 256
    else:
        raise ValueError(
            f'size must be one of "small", "medium", "large", got {size}')

    input_layer = tf.keras.layers.Input(input_shape)
    for i in range(num_hidden):
        if i == 0:
            x = tf.keras.layers.Dense(
                hidden_size, activation='relu')(input_layer)
        else:
            x = tf.keras.layers.Dense(hidden_size, activation='relu')(x)
    output_layer = tf.keras.layers.Dense(
        num_outputs, activation=output_activation)(x)

    return tf.keras.models.Model(input_layer, output_layer)


def mimic_model(
    trained_model: BaseEstimator,
    nnet: tf.keras.models.Model,
    training_data: np.ndarray,
    test_data: np.ndarray,
    test_labels: np.ndarray,
    problem_type: str,
    loss: str,
    metrics: Union[str, list],
    optimizer: str,
    mimic_proba: bool = False,
    retention: float = 0.9,
    batch_size: int = 32,
    epochs: int = 100,
    starting_sparsification: int = 0,
    max_sparsification: int = 99,
    sparsification_rate: int = 5
) -> tf.keras.models.Model:
    """
    Train a sparse neural network to mimic a scikit-learn model

    Parameters
    ----------
    trained_model : sklearn model
        The model that is already trained
    nnet : TensorFlow keras Model
        The neural network to train to mimic the trained model
    training_data : array or array-like
        The input data that was used to train the trained model
    test_data : array or array-like
        The input data to be used for testing
    test_labels : array or array-like
        The output data used in testing
    problem_type : str
        The type of problem, either 'classification' or 'regression'
    loss : str or keras loss function
        The loss to use
    metrics : str, function or list of str, function
        Metrics to measure
    optimizer : str or keras optimizer
        The optimizer to use
    mimic_proba : bool (default False)
        For classification, mimic the probability outputs
    retention : float (default 0.9)
        The retention of performance to allow further pruning
    batch_size : int (default 32)
        The batch size to use while training
    epochs : int (default 100)
        The number of epochs (if early stopping is not met beforehand)
    starting_sparsification : int (default 0)
        The starting model sparsification
    max_sparsification : int (default 99)
        The maximum sparsification to allow
    sparsification_rate : int (default 5)
        The sparsification rate when invoked

    Returns
    -------
    nnet : TensorFlow keras Model
        The trained model
    """

    # Check problem type
    if problem_type not in ['classification', 'regression']:
        raise ValueError(
            'problem_type must be "classification" or "regression"')

    # Check that the model to mimic is a sklearn model
    if not isinstance(trained_model, BaseEstimator):
        raise TypeError(
            'Currently can only convert trained scikit-learn models')

    # Get the predictions
    if isinstance(trained_model, TransformerMixin):
        training_predictions = trained_model.transform(training_data)
        testing_predictions = trained_model.transform(test_data)
    elif isinstance(trained_model, (ClassifierMixin, RegressorMixin)):
        if isinstance(trained_model, ClassifierMixin) and mimic_proba:
            training_predictions = trained_model.predict_proba(training_data)
            testing_predictions = trained_model.predict(test_data)
        else:
            training_predictions = trained_model.predict(training_data)
            testing_predictions = trained_model.predict(test_data)
    else:
        raise TypeError(
            'trained_model is not a scikit-learn TransformerMixin, ClassifierMixin, or RegressorMixin')

    if problem_type == 'classification':
        cutoff = retention
    else:
        cutoff = mean_squared_error(
            test_labels, testing_predictions, squared=False) / retention

    # Add layer masks
    nnet = beyondml.tflow.utils.add_layer_masks(nnet)
    nnet = beyondml.tflow.utils.train_model(
        nnet,
        training_data,
        training_predictions,
        loss,
        metrics,
        optimizer,
        cutoff,
        batch_size,
        epochs,
        starting_sparsification,
        max_sparsification,
        sparsification_rate
    )

    nnet = beyondml.tflow.utils.remove_layer_masks(nnet)
    nnet_preds = nnet.predict(test_data)
    _print_report(test_labels, testing_predictions, nnet_preds, problem_type)
    return nnet


def get_model(
    model_type: str,
    input_shape: Union[int, tuple],
    num_outputs: int,
    output_activation: str,
    size: str = 'small',
    vocab_size: Union[None, int] = None
):
    """
    Get a pre-configured model for different use cases

    Parameters
    ----------
    model_type : str
        Either 'cv', 'nlp_embedding', or 'fc', defining the model type
    input_shape : int or tuple of int
        The input shape to the model
    num_outputs : int
        The output shape of the model
    output_activation : str or keras activation function
        The activation of the final layer of the model
    size : str (default 'small')
        One of either 'small', 'medium', or 'large'
    vocab_size : str or None (default None)
        Size of the vocab, if model_type is 'nlp_embedding'

    Returns
    -------
    model : TensorFlow Keras model
        The model
    """
    if model_type == 'cv':
        return _get_cv_model(size, input_shape, num_outputs, output_activation)
    elif model_type == 'nlp_embedding':
        return _get_embedding_model(size, vocab_size, input_shape, num_outputs, output_activation)
    elif model_type == 'fc':
        return _get_fc_model(size, input_shape, num_outputs, output_activation)
    else:
        raise ValueError(
            f'model_type must be one of "cv", "nlp_embedding", or "fc", got {model_type}')



"""
The aisquared.config.rendering subpackage contains objects for configuring how rendering of predictions is to occur.
"""

from .ImageRendering import ImageRendering
from .ObjectRendering import ObjectRendering
from .DocumentRendering import DocumentRendering
from .WordRendering import WordRendering
from .FilterRendering import FilterRendering
from .ContainerRendering import ContainerRendering
from .DoughnutChartRendering import DoughnutChartRendering
from .TableRendering import TableRendering
from .HTMLTagRendering import HTMLTagRendering
from .BarChartRendering import BarChartRendering
from .LineChartRendering import LineChartRendering
from .DashboardReplacementRendering import DashboardReplacementRendering
from .PieChartRendering import PieChartRendering
from .SOSRendering import SOSRendering
from .TextRendering import TextRendering
from .CustomRendering import CustomRendering
from .ChatRendering import ChatRendering

# from .DashboardRendering import DashboardRendering

from .utils import save_default_css



from aisquared.base import BaseObject


class DashboardReplacementRendering(BaseObject):
    """
    Rendering for dashboard replacement

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.rendering.DashboardReplacementRendering(
        'test_anchor_selector'
    )
    >>> my_obj.to_dict()
    {'className': 'DashboardReplacementRendering',
    'label': '',
    'params': {'anchorSelector': 'test_anchor_selector', 'whereReplace': ''}}

    """

    def __init__(
        self,
        anchor_selector: str,
        where_replace: str = '',
        label: str = ''
    ):
        """
        Parameters
        ----------
        anchor_selector : str
            The anchor selector
        where_replace : str (default '')
            Where to replace
        label : str (default '')
            The label for the rendering
        """
        super().__init__()
        self.anchor_selector = anchor_selector
        self.where_replace = where_replace
        self.label = label

    @property
    def anchor_selector(self):
        return self._anchor_selector

    @anchor_selector.setter
    def anchor_selector(self, value):
        self._anchor_selector = value

    @property
    def where_replace(self):
        return self._where_replace

    @where_replace.setter
    def where_replace(self, value):
        self._where_replace = value

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        self._label = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'DashboardReplacementRendering',
            'label': self.label,
            'params': {
                'anchorSelector': self.anchor_selector,
                'whereReplace': self.where_replace
            }
        }



import pytest
import aisquared


def test_binary_feedback():
    with pytest.raises(TypeError):
        aisquared.config.feedback.BinaryFeedback()
    with pytest.raises(ValueError):
        aisquared.config.feedback.BinaryFeedback(['foo', 'bar', 'baz'])

    feedback = aisquared.config.feedback.BinaryFeedback(
        ['foo', 'bar']
    )
    assert feedback.to_dict() == {
        'className': 'BinaryFeedback',
        'params': {
            'labelMap': ['foo', 'bar']
        }
    }


def test_multiclass_feedback():
    feedback = aisquared.config.feedback.MulticlassFeedback(
        ['foo', 'bar', 'baz']
    )
    assert feedback.to_dict() == {
        'className': 'MulticlassFeedback',
        'params': {
            'labelMap': ['foo', 'bar', 'baz']
        }
    }


def test_regression_feedback():
    feedback = aisquared.config.feedback.RegressionFeedback()
    assert feedback.to_dict() == {
        'className': 'RegressionFeedback',
        'params': dict()
    }


def test_simple_feedback():
    feedback = aisquared.config.feedback.SimpleFeedback()
    assert feedback.to_dict() == {
        'className': 'SimpleFeedback',
        'params': dict()
    }


def test_qualitative_feedback():
    feedback = aisquared.config.feedback.QualitativeFeedback()

    with pytest.raises(ValueError):
        feedback.add_question(
            'test',
            'test'
        )
        feedback.add_question(
            'test',
            'singleChoice'
        )

    feedback.add_question(
        'test',
        'singleChoice',
        ['foo', 'bar', 'baz']
    )

    feedback.add_question(
        'test',
        'multiChoice',
        ['foo', 'bar', 'baz']
    )

    feedback.add_question(
        'test',
        'text'
    )

    assert feedback.to_dict() == {
        'className': 'QualitativeFeedback',
        'params': {
            'questions': [
                {
                    'question': 'test',
                    'answerType': 'singleChoice',
                    'choices': ['foo', 'bar', 'baz']
                },
                {
                    'question': 'test',
                    'answerType': 'multiChoice',
                    'choices': ['foo', 'bar', 'baz']
                },
                {
                    'question': 'test',
                    'answerType': 'text'
                }
            ]
        }
    }


def test_model_feedback():
    feedback = aisquared.config.feedback.ModelFeedback()

    with pytest.raises(ValueError):
        feedback.add_question(
            'test',
            'test'
        )
        feedback.add_question(
            'test',
            'singleChoice'
        )

    feedback.add_question(
        'test',
        'singleChoice',
        ['foo', 'bar', 'baz']
    )

    feedback.add_question(
        'test',
        'multiChoice',
        ['foo', 'bar', 'baz']
    )

    feedback.add_question(
        'test',
        'text'
    )

    assert feedback.to_dict() == {
        'className': 'ModelFeedback',
        'params': {
            'questions': [
                {
                    'question': 'test',
                    'answerType': 'singleChoice',
                    'choices': ['foo', 'bar', 'baz']
                },
                {
                    'question': 'test',
                    'answerType': 'multiChoice',
                    'choices': ['foo', 'bar', 'baz']
                },
                {
                    'question': 'test',
                    'answerType': 'text'
                }
            ]
        }
    }



"""
Some allowed configuration parameters - not meant to be directly called by the end user
"""

LOCATIONS = [
    None,
    'top',
    'bottom',
    'left',
    'right',
    'topright',
    'topleft',
    'bottomright',
    'bottomleft'
]

COLORS = [
    'red',
    'orange',
    'green',
    'purple',
    'yellow',
    'pink',
    'black',
    'grey',
    'white',
    'blue'
]

BADGES = [
    'circle',
    'square',
    'star',
    'underline'
]

WORD_LISTS = [
    'input',
    'result'
]

QUALIFIERS = [
    'gt',
    'lt',
    'gte',
    'lte',
    'in',
    'ne',
    'nin'
]

POSITIONS = [
    'absolute',
    'static'
]

STATIC_POSITIONS = [
    'append',
    'prepend'
]



"""
The aisquared.config.preprocessing subpackage contains utilities to configure the preprocessing of data in the data pipeline.  It contains
 three separate subpackages, aisquared.config.preprocessing.text, aisquared.config.preprocessing.image, and aisquared.config.preprocessing.tabular, which configure the preprocessing of different types of data.
"""



from aisquared.base import BaseObject


class DeployedModel(BaseObject):
    """
    Interaction with a remote model

    Example usage:

    >>> import aisquared
    >>> analytic = aisquared.config.analytic.DeployedModel(
        'model_url',
        'text'
    )
    >>> analytic.to_dict()
    {'className': 'DeployedModel',
    'params': {'url': 'model_url',
    'inputType': 'text',
    'headers': None,
    'bodyKey': None,
    'returnKey': None,
    'bodySetup': None,
    'bodySetupReplaceValue': None
    }}

    """

    def __init__(
        self,
        url: str,
        input_type: str,
        headers: dict = None,
        body_key: str = None,
        return_key: str = None,
        body_setup: dict = None,
        body_setup_replace_value: str = '{DATAPOINT}'
    ):
        """
        Parameters
        ----------
        url : str
            The base URL for the remote endpoint
        input_type : str
            The input type to the model. Either one of 'cv', 'text', or 'tabular'
        headers : dict or None (default None)
            Headers to use when calling the endpoint
        body_key : str or None (default None)
            The key to use for the data to be sent to the endpoint
        return_key : str or None (default None)
            The key to parse the returned value from
        body_setup : dict or None (default None)
            The prototype for the way the body of the request to the API is set up
        body_setup_replace_value : str (default '{DATAPOINT}')
            The value in the stringified version of the `body_setup` that gets replaced by the harvested data
        """
        super().__init__()
        self.url = url
        self.input_type = input_type
        self.headers = headers
        self.body_key = body_key
        self.return_key = return_key
        self.body_setup = body_setup
        self.body_setup_replace_value = body_setup_replace_value

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

    @property
    def input_type(self):
        return self._input_type

    @input_type.setter
    def input_type(self, value):
        self._input_type = value

    @property
    def headers(self):
        return self._headers

    @headers.setter
    def headers(self, value):
        self._headers = value

    @property
    def body_key(self):
        return self._body_key

    @body_key.setter
    def body_key(self, value):
        self._body_key = value

    @property
    def return_key(self):
        return self._return_key

    @return_key.setter
    def return_key(self, value):
        self._return_key = value

    @property
    def body_setup(self):
        return self._body_setup

    @body_setup.setter
    def body_setup(self, value):
        self._body_setup = value

    @property
    def body_setup_replace_value(self):
        return self._body_setup_replace_value

    @body_setup_replace_value.setter
    def body_setup_replace_value(self, value):
        self._body_setup_replace_value = value

    def to_dict(self) -> dict:
        """
        Get the config object as a dictionary
        """
        return {
            'className': 'DeployedModel',
            'params': {
                'url': self.url,
                'inputType': self.input_type,
                'headers': self.headers,
                'bodyKey': self.body_key,
                'returnKey': self.return_key,
                'bodySetup': self.body_setup,
                'bodySetupReplaceValue': self.body_setup_replace_value
            }
        }



from aisquared.base import BaseObject

ALLOWED_TYPES = ['singleChoice', 'multiChoice', 'text']
CHOICE_TYPES = ALLOWED_TYPES[:2]


def _create_question_dict(question, answer_type, choices):
    if answer_type in CHOICE_TYPES:
        return {
            'question': question,
            'answerType': answer_type,
            'choices': choices
        }

    # Since already checked for CHOICE_TYPES, must be 'text'
    elif answer_type in ALLOWED_TYPES:
        return {
            'question': question,
            'answerType': answer_type
        }
    else:
        raise ValueError(f'answer_type must be one of {ALLOWED_TYPES}')


class ModelFeedback(BaseObject):
    """
    Feedback object for questions and answers for an individual model.

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.feedback.ModelFeedback()
    >>> my_obj.add_question(
        'How is the model performing?',
        choices = ['very poorly', 'poorly', 'neutral', 'well', 'very well']
    )
    >>> my_obj.add_question(
        'Any additional feedback?',
        'text'
    )
    >>> my_obj.to_dict()
    {'className': 'ModelFeedback',
    'params': {'questions': [{'question': 'How is the model performing?',
    'answerType': 'singleChoice',
    'choices': ['very poorly', 'poorly', 'neutral', 'well', 'very well']},
    {'question': 'Any additional feedback?', 'answerType': 'text'}]}}
    """

    def __init__(self):
        super().__init__()
        self.questions = []

    def add_question(
        self,
        question: str,
        answer_type: str = 'singleChoice',
        choices: list = []
    ):
        """
        Add a question to be asked.

        Parameters
        ----------
        question : str
            The question to be asked.
        answer_type : str (default 'singleChoice')
            One of either 'singleChoice', 'multiChoice', or 'text'
        choices : list (default [])
            The choices to be provided, if `answer_type` is 'singleChoice' or 'multiChoice'
        """
        if answer_type in ['singleChoice', 'multiChoice'] and choices == []:
            raise ValueError(
                'If singleChoice or multiChoice is indicated, choices must be provided')

        if answer_type not in ['singleChoice', 'multiChoice', 'text']:
            raise ValueError(
                'answer_type must be one of "singleChoice", "multiChoice", or "text"')

        self.questions.append(_create_question_dict(
            question, answer_type, choices))

    def to_dict(self) -> dict:
        """
        Return the object as a dictionary
        """
        return {
            'className': 'ModelFeedback',
            'params': {
                'questions': self.questions
            }
        }



"""
The aisquared.config.analytic subpackage contains objects for packaging individual analytics.
"""

from .DeployedAnalytic import DeployedAnalytic
from .DeployedModel import DeployedModel
from .LocalModel import LocalModel
from .LocalAnalytic import LocalAnalytic
from .ReverseMLWorkflow import ReverseMLWorkflow
from .OnnxModel import OnnxModel



from typing import Union
from aisquared.base import BaseObject, ALLOWED_STAGES, HARVESTING_CLASSES, PREPROCESSING_CLASSES, ANALYTIC_CLASSES, POSTPROCESSING_CLASSES, RENDERING_CLASSES, FEEDBACK_CLASSES, LOCAL_CLASSES
import tensorflowjs as tfjs
import tensorflow as tf
import shutil
import json
import os


class ModelConfiguration(BaseObject):
    """
    Configuration object for deploying a model or analytic
    """

    def __init__(
            self,
            name: str,
            harvesting_steps: Union[BaseObject, list] = None,
            preprocessing_steps: Union[BaseObject, list] = None,
            analytic: Union[BaseObject, list] = None,
            postprocessing_steps: Union[BaseObject, list] = None,
            rendering_steps: Union[BaseObject, list] = None,
            feedback_steps: Union[BaseObject, list] = None,
            stage: str = ALLOWED_STAGES[0],
            version: int = None,
            description: str = '',
            mlflow_uri: str = None,
            mlflow_user: str = None,
            mlflow_token: str = None,
            owner: str = None,
            url: str = '*',
            auto_run: bool = False,
            documentation_link: str = '',
            warnings: list = None
    ):
        """
        Parameters
        ----------
        name : str
            The name of the deployed analytic
        harvesting_steps : None, Harvesting object or list of Harvesting objects
            Harvesters to use with the analytic
        preprocessing_steps : Preprocessing object or list of Preprocessing objects or None
            Preprocessers to use
        analytic : Analytic object or list of Analytic objects
            Analytics to use
        postprocessing_steps : Postprocessing object or list of Postprocessing objects or None
            Postprocessers to use
        rendering_steps : Rendering object or list of Rendering objects or None
            Renderers to use
        feedback_steps : None, Feedback object or list of Feedback objects or None (default None)
            Feedback steps to use
        stage : str (default 'experimental')
            The stage of the model, from 'experimental', 'staging', 'production'
        version : str or None (default None)
            Version of the analytic
        description : str (default '')
            The description of the analytic
        mlflow_uri : str or None (default None)
            MLFlow URI to use, if applicable
        mlflow_user : str or None (default None)
            MLFlow user to use, if applicable
        mlflow_token : str or None (default None)
            MLFlow token to use, if applicable
        owner : str or None (default None)
            The owner of the model
        url : str (default '*')
            URL or URL pattern to match
        auto_run : bool (default False)
            Whether to automatically run this file when on a valid page
        documentation_link : str (default '')
            If provided, a URL to the model card for the .air file
        warnings : list or None (default None)
            Any warnings to be shown when the model runs
        """
        super().__init__()
        self.name = name
        self.harvesting_steps = harvesting_steps
        self.preprocessing_steps = preprocessing_steps
        self.analytic = analytic
        self.postprocessing_steps = postprocessing_steps
        self.rendering_steps = rendering_steps
        self.stage = stage
        self.feedback_steps = feedback_steps
        self.version = version
        self.description = description
        self.mlflow_uri = mlflow_uri
        self.mlflow_user = mlflow_user
        self.mlflow_token = mlflow_token
        self.owner = owner
        self.url = url
        self.auto_run = auto_run
        self.documentation_link = documentation_link
        self.warnings = warnings

    # name
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = str(value)

    # harvesting_steps
    @property
    def harvesting_steps(self):
        return self._harvesting_steps

    @harvesting_steps.setter
    def harvesting_steps(self, value):
        harvesting_classes = HARVESTING_CLASSES + (ModelConfiguration,)
        if value is None or (isinstance(value, list) and all([val is None for val in value])):
            self._harvesting_steps = value
        elif isinstance(value, harvesting_classes):
            self._harvesting_steps = [value]
        elif isinstance(value, list) and all([isinstance(val, harvesting_classes) for val in value]):
            self._harvesting_steps = value
        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, harvesting_classes) for val in value for v in val]):
            self._harvesting_steps = value
        else:
            raise ValueError(
                'harvesting_steps must be a None, single Harvester object, a list of Harvester objects, or a list of list of harvester objects')

    # preprocessing_steps
    @property
    def preprocessing_steps(self):
        return self._preprocessing_steps

    @preprocessing_steps.setter
    def preprocessing_steps(self, value):
        if value is None:
            self._preprocessing_steps = value
        elif isinstance(value, PREPROCESSING_CLASSES):
            self._preprocessing_steps = [value]
        elif isinstance(value, list) and all([isinstance(val, PREPROCESSING_CLASSES) for val in value]):
            self._preprocessing_steps = value
        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, PREPROCESSING_CLASSES) for val in value for v in val]):
            self._preprocessing_steps = value
        elif value is None:
            self._preprocessing_steps = value
        else:
            raise ValueError(
                'preprocessing_steps must a single Preprocesser object, a list of Preprocesser objects, or a list of list of preprocesser objects')

    # analytic
    @property
    def analytic(self):
        return self._analytic

    @analytic.setter
    def analytic(self, value):
        if value is None:
            self._analytic = value
        elif isinstance(value, ANALYTIC_CLASSES):
            self._analytic = [value]
        elif isinstance(value, list) and all([isinstance(val, ANALYTIC_CLASSES) for val in value]):
            self._analytic = value
        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, ANALYTIC_CLASSES) for val in value for v in val]):
            self._analytic = value
        else:
            raise ValueError(
                'analytic must be a single Analytic object, a list of Analytic objects, or a list of list of Analtyic objects')

    # postprocessing_steps
    @property
    def postprocessing_steps(self):
        return self._postprocessing_steps

    @postprocessing_steps.setter
    def postprocessing_steps(self, value):
        if value is None:
            self._postprocessing_steps = value
        elif isinstance(value, POSTPROCESSING_CLASSES):
            self._postprocessing_steps = [value]
        elif isinstance(value, list) and all([isinstance(val, POSTPROCESSING_CLASSES) for val in value]):
            self._postprocessing_steps = value
        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, POSTPROCESSING_CLASSES) for val in value for v in val]):
            self._postprocessing_steps = value
        elif value is None:
            self._postprocessing_steps = value
        else:
            raise ValueError(
                'postprocessing_steps must be a single Postprocessing object, a list of Postprocessing objects, or a list of list of Postprocessing objects')

    # rendering_steps
    @property
    def rendering_steps(self):
        return self._rendering_steps

    @rendering_steps.setter
    def rendering_steps(self, value):
        if isinstance(value, RENDERING_CLASSES) or value is None:
            self._rendering_steps = [value]
        elif isinstance(value, list) and all([isinstance(val, RENDERING_CLASSES) for val in value]):
            self._rendering_steps = value
        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, RENDERING_CLASSES) for val in value for v in val]):
            self._rendering_steps = value
        else:
            raise ValueError(
                'rendering_steps must be a single Rendering object, a list of Rendering objects, a list of list of Rendering objects, or None')

    # feedback_steps
    @property
    def feedback_steps(self):
        return self._feedback_steps

    @feedback_steps.setter
    def feedback_steps(self, value):
        if value is None:
            self._feedback_steps = value
        elif isinstance(value, FEEDBACK_CLASSES):
            self._feedback_steps = [value]
        elif isinstance(value, list) and all([isinstance(val, FEEDBACK_CLASSES) for val in value]):
            self._feedback_steps = value
        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, FEEDBACK_CLASSES) for val in value for v in val]):
            self._feedback_steps = value
        else:
            raise ValueError(
                'feedback_steps must be a single Feedback object, a list of Feedback objects, a list of list of Feedback objects, or None')

    # stage
    @property
    def stage(self):
        return self._stage

    @stage.setter
    def stage(self, value):
        if value not in ALLOWED_STAGES:
            raise ValueError(f'stage must be one of {ALLOWED_STAGES}')
        self._stage = value

    # version
    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, value):
        self._version = str(value) if value is not None else value

    # description
    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = str(value)

    # mlflow_uri
    @property
    def mlflow_uri(self):
        return self._mlflow_uri

    @mlflow_uri.setter
    def mlflow_uri(self, value):
        self._mlflow_uri = value

    # mlflow_user
    @property
    def mlflow_user(self):
        return self._mlflow_user

    @mlflow_user.setter
    def mlflow_user(self, value):
        self._mlflow_user = value

    # mlflow_token
    @property
    def mlflow_token(self):
        return self._mlflow_token

    @mlflow_token.setter
    def mlflow_token(self, value):
        self._mlflow_token = value

    # owner
    @property
    def owner(self):
        return self._owner

    @owner.setter
    def owner(self, value):
        if not isinstance(value, str) and value is not None:
            raise ValueError('owner must be a string or None')
        self._owner = value

    # url
    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        if not isinstance(value, str):
            raise ValueError('url must be string')
        self._url = value

    # auto_run
    @property
    def auto_run(self):
        return self._auto_run

    @auto_run.setter
    def auto_run(self, value):
        if not isinstance(value, bool):
            raise TypeError('auto_run must be Boolean valued')
        self._auto_run = str(value).lower()

    # documentation_link
    @property
    def documentation_link(self):
        return self._documentation_link

    @documentation_link.setter
    def documentation_link(self, value):
        if not isinstance(value, str):
            raise TypeError('documentation_link must be str')
        self._documentation_link = value

    @property
    def warnings(self):
        return self._warnings

    @warnings.setter
    def warnings(self, value):
        if value is None:
            self._warnings = value
        else:
            if isinstance(value, str):
                value = [value]
            if not all([isinstance(val, str) for val in value]):
                raise TypeError('warnings must all be strings')
            self._warnings = value

    # harvester_dict
    @property
    def harvester_dict(self):
        harvesting_classes = HARVESTING_CLASSES + (ModelConfiguration,)
        if self.harvesting_steps is None or (isinstance(self.harvesting_steps, list) and all([val is None for val in self.harvesting_steps])):
            return None
        elif isinstance(self.harvesting_steps, list) and all([isinstance(val, harvesting_classes) for val in self.harvesting_steps]):
            return [val.to_dict() for val in self.harvesting_steps]
        else:
            return [
                [v.to_dict() for v in val] for val in self.harvesting_steps
            ]

    # preprocessing_dict
    @property
    def preprocesser_dict(self):
        if self.preprocessing_steps is None:
            return self.preprocessing_steps
        elif isinstance(self.preprocessing_steps, list) and all([isinstance(val, PREPROCESSING_CLASSES) for val in self.preprocessing_steps]):
            return [val.to_dict() for val in self.preprocessing_steps]
        else:
            return [
                [v.to_dict() for v in val] for val in self.preprocessing_steps
            ]

    # analytic dict
    @property
    def analytic_dict(self):
        if isinstance(self.analytic, list) and all([isinstance(val, ANALYTIC_CLASSES) for val in self.analytic]):
            return [val.to_dict() for val in self.analytic]
        else:
            return [
                [v.to_dict() for v in val] for val in self.analytic
            ]

    # postprocesser_dict
    @property
    def postprocesser_dict(self):
        if self.postprocessing_steps is None:
            return self.postprocessing_steps
        elif isinstance(self.postprocessing_steps, list) and all([isinstance(val, POSTPROCESSING_CLASSES) for val in self.postprocessing_steps]):
            return [val.to_dict() for val in self.postprocessing_steps]
        else:
            return [
                [v.to_dict() for v in val] for val in self.postprocessing_steps
            ]

    # render_dict
    @property
    def render_dict(self):
        if self.rendering_steps[0] is None:
            render_list = []
        elif isinstance(self.rendering_steps, list) and all([isinstance(val, RENDERING_CLASSES) for val in self.rendering_steps]):
            render_list = [val.to_dict() for val in self.rendering_steps]
        else:
            render_list = [
                [v.to_dict() for v in val] for val in self.rendering_steps
            ]
        to_ret = []
        for val in render_list:
            if isinstance(val, list):
                for v in val:
                    to_ret.append(v)
            else:
                to_ret.append(val)
        return to_ret

    # feedback_dict

    @property
    def feedback_dict(self):
        if self.feedback_steps is None:
            return self.feedback_steps
        elif isinstance(self.feedback_steps, list) and all([isinstance(val, FEEDBACK_CLASSES) for val in self.feedback_steps]):
            return [val.to_dict() for val in self.feedback_steps]
        else:
            return [
                [v.to_dict() for v in val] for val in self.feedback_steps
            ]

    def get_model_filenames(self) -> list:
        """
        Get filenames for all models in the configuration
        """
        filenames = []
        if self.harvesting_steps is None or (isinstance(self.harvesting_steps, list) and all([val is None for val in self.harvesting_steps])):
            harvesting_list = []
        elif isinstance(self.harvesting_steps[0], list):
            harvesting_list = [
                h for harvester in self.harvesting_steps for h in harvester]
        else:
            harvesting_list = self.harvesting_steps
        for harvester in harvesting_list:
            if isinstance(harvester, ModelConfiguration):
                filenames.extend(harvester.get_model_filenames())

        if isinstance(self.analytic[0], ANALYTIC_CLASSES):
            for a in self.analytic:
                if isinstance(a, LOCAL_CLASSES):
                    try:
                        filenames.append(a.path)
                    except Exception:
                        pass
        else:
            for analytic in self.analytic:
                for a in analytic:
                    if isinstance(a, LOCAL_CLASSES):
                        try:
                            filenames.append(a.path)
                        except Exception:
                            pass
        return [f for f in filenames if f is not None]

    def to_dict(self) -> dict:
        """
        Get the object as a dictionary
        """
        return {
            'className': 'ModelConfiguration',
            'params': {
                'name': self.name,
                'harvestingSteps': self.harvester_dict,
                'preprocessingSteps': self.preprocesser_dict,
                'analytics': self.analytic_dict,
                'postprocessingSteps': self.postprocesser_dict,
                'renderingSteps': self.render_dict,
                'feedbackSteps': self.feedback_dict,
                'stage': self.stage,
                'version': self.version,
                'description': self.description,
                'mlflowUri': self.mlflow_uri,
                'mlflowUser': self.mlflow_user,
                'mlflowToken': self.mlflow_token,
                'owner': self.owner,
                'url': self.url,
                'autoRun': self.auto_run,
                'documentURL': self.documentation_link,
                'warnings': self.warnings
            }
        }

    def compile(self, filename: str = None, dtype: str = None) -> None:
        """
        Compile the object into a '.air' file, which can then be dragged and
        dropped into applications using the AI Squared JavaScript SDK

        Parameters
        ----------
        filename : path-like or None (default None)
            Filename to compile to. If None, defaults to '{NAME}.air', where {NAME} is the
            name of the analytic
        dtype : str or None (default None)
            The datatype to use for the model weights. If None, defaults to 'float32'
        """
        if filename is None:
            filename = self.name + '.air'

        if dtype is None:
            dtype_map = None
        else:
            dtype_map = {dtype: '*'}

        dirname = os.path.join('.', os.path.splitext(filename)[0])

        # write the object as json config
        os.makedirs(dirname, exist_ok=True)
        with open(os.path.join(dirname, 'config.json'), 'w') as f:
            json.dump(self.to_dict(), f)

        # go through the files and copy them/make them tfjs files
        filenames = self.get_model_filenames()
        if len(filenames) != 0:
            for f in filenames:
                if os.path.splitext(f)[-1] in ['.h5', '.keras']:
                    model = tf.keras.models.load_model(f)
                    model_dir = os.path.join(dirname, os.path.split(f)[-1])
                    tfjs.converters.save_keras_model(
                        model, model_dir, quantization_dtype_map=dtype_map)
                else:
                    if os.path.isdir(f):
                        shutil.copytree(f, os.path.join(dirname, f))
                    else:
                        shutil.copy(f, dirname)

        # go through the entire directory of dirname, grab all files, and make
        # the archive file
        shutil.make_archive(filename, 'zip', dirname)

        # Move the archive file to just have .air
        shutil.move(filename + '.zip', filename)

        # Remove the temp directory
        shutil.rmtree(dirname, ignore_errors=True)



from .AISquaredAPIException import AISquaredAPIException
from .NoResultsFoundError import NoResultsFoundError
from aisquared.base import ENDPOINTS
import pandas as pd
import requests


def _test_connection(
        url: str
):
    url = f'{url}/{ENDPOINTS["health"]}'

    with requests.Session() as sess:
        resp = sess.get(
            url
        )

    if resp.ok:
        return resp.ok
    else:
        raise AISquaredAPIException(
            f'Connection could not be established. Status code {resp.status_code}')


def _check_results_length(
        df: pd.DataFrame
):
    if df.shape[0] == 0:
        return NoResultsFoundError('No results found.')



from aisquared.base import BaseObject


class BinaryFeedback(BaseObject):
    """
    Feedback for binary classification

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.feedback.BinaryFeedback(['class1', 'class2'])
    >>> my_obj.to_dict()
    {'className': 'BinaryFeedback', 'params': {'labelMap': ['class1', 'class2']}}

    """

    def __init__(self, label_map: list):
        """
        Parameters
        ----------
        label_map : list of two values
            The two values to map to
        """
        super().__init__()
        self.label_map = label_map

    @property
    def label_map(self):
        return self._label_map

    @label_map.setter
    def label_map(self, value):
        if not isinstance(value, list) or len(value) != 2:
            raise ValueError('label_map must be list of length 2')
        self._label_map = value

    def to_dict(self) -> dict:
        """
        Return the object as a dictionary
        """
        return {
            'className': 'BinaryFeedback',
            'params': {
                'labelMap': self.label_map
            }
        }



from aisquared.base import BaseObject


class SOSRendering(BaseObject):
    """
    Rendering of an SOS dashboard

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.rendering.SOSRendering(True)
    >>> my_obj.to_dict()
    {'className': 'SOSRendering', 'label': '', 'params': {'canToggle': True}}

    """

    def __init__(
        self,
        can_toggle: bool,
        label: str = ''
    ):
        """
        Parameters
        ----------
        can_toggle : bool
            Whether the dashboard can be toggled on and off
        label : str (default '')
            The label for the rendering
        """
        super().__init__()
        self.can_toggle = can_toggle
        self.label = label

    @property
    def can_toggle(self):
        return self._can_toggle

    @can_toggle.setter
    def can_toggle(self, value):
        if not isinstance(value, bool) and value is not None:
            raise TypeError('can_toggle must be boolean or None')
        self._can_toggle = value

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        if not isinstance(value, str):
            raise TypeError('label must be str')
        self._label = value

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'SOSRendering',
            'label': self.label,
            'params': {
                'canToggle': self.can_toggle
            }
        }



import pytest
import aisquared


def test_base_notimplemented():
    obj = aisquared.base.BaseObject()
    with pytest.raises(NotImplementedError):
        obj.to_dict()


def test_stages():
    assert aisquared.base.ALLOWED_STAGES == [
        'experimental',
        'staging',
        'production'
    ]


def test_locations():
    assert aisquared.base.LOCATIONS == [
        None,
        'top',
        'bottom',
        'left',
        'right',
        'topright',
        'topleft',
        'bottomright',
        'bottomleft'
    ]


def test_badges():
    assert aisquared.base.BADGES == [
        'circle',
        'square',
        'star',
        'underline'
    ]


def test_word_lists():
    assert aisquared.base.WORD_LISTS == [
        'input',
        'result'
    ]


def test_qualifiers():
    assert aisquared.base.QUALIFIERS == [
        'gt',
        'lt',
        'gte',
        'lte',
        'in',
        'ne',
        'nin'
    ]



from typing import Union
from aisquared.base import BaseObject, QUALIFIERS

_ALLOWED_SOURCES = [
    'inputs',
    'outputs'
]


class FilterRendering(BaseObject):
    """
    Object which dictates how predictions are to be passed to downstream analytics

    Example usage:

    >>> import aisquared
    >>> my_obj = aisquared.config.rendering.FilterRendering(
        'inputs',
        'key',
        'gt',
        0.2
    )
    >>> my_obj.to_dict()
    {'className': 'FilterRendering',
    'params': {'source': 'inputs', 'key': 'key', 'qualifier': 'gt', 'value': 0.2}}

    """

    def __init__(
        self,
        source: str,
        key: str,
        qualifier: str,
        value: Union[list, str, int, float]
    ):
        """
        Parameters
        ----------
        source : either one of 'inputs', 'outputs'
            The source to look for the key value
        key : str
            The key to use for filtering
        qualifier : one of 'gt', 'lt', 'gte', 'lte', 'in', 'ne', 'nin'
            Qualifier to use for filtering
        value : list, string, or numeric
            The value to use to filter value on
        """
        super().__init__()
        self.source = source
        self.key = key
        self.qualifier = qualifier
        self.value = value

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, value):
        if value not in _ALLOWED_SOURCES:
            raise ValueError(f'source must be one of {_ALLOWED_SOURCES}')
        self._source = value

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, value):
        if not isinstance(value, str):
            raise TypeError('key must be string')
        self._key = value

    @property
    def qualifier(self):
        return self._qualifier

    @qualifier.setter
    def qualifier(self, value):
        if value not in QUALIFIERS:
            raise ValueError(f'qualifier must be one of {QUALIFIERS}')
        self._qualifier = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        if not isinstance(val, (list, str, int, float)):
            raise TypeError('value must be list, string, or numeric')
        self._value = val

    def to_dict(self) -> dict:
        """
        Get the configuration object as a dictionary
        """
        return {
            'className': 'FilterRendering',
            'params': {
                'source': self.source,
                'key': self.key,
                'qualifier': self.qualifier,
                'value': self.value
            }
        }



"""
The aisquared.config.harvesting subpackage contains objects for configuring harvesting of data.
"""

from .ImageHarvester import ImageHarvester
from .TextHarvester import TextHarvester
from .InputHarvester import InputHarvester
from .QueryParameterHarvester import QueryParameterHarvester
from .ChatbotHarvester import ChatbotHarvester



from typing import Union
from aisquared.base import BaseObject, ALLOWED_STAGES
from .CustomObject import CustomObject
import tensorflowjs as tfjs
import tensorflow as tf
import shutil
import json
import os

LOCAL_CLASSES = ['LocalModel', 'LocalAnalytic', 'CustomObject']


class GraphConfiguration(BaseObject):
    """
    Configuration object for deploying a set of processing steps and/or analytics as a dependency graph
    """

    def __init__(
            self,
            name: str,
            stage: str = ALLOWED_STAGES[0],
            version: int = None,
            description: str = '',
            mlflow_uri: str = None,
            mlflow_user: str = None,
            mlflow_token: str = None,
            owner: str = None,
            url: str = '*',
            auto_run: bool = False,
            documentation_link: str = ''
    ):
        """
        Parameters
        ----------
        name : str
            The name of the deployed analytic
        stage : str (default 'experimental')
            The stage of the model, from 'experimental', 'staging', 'production'
        version : str or None (default None)
            Version of the analytic
        description : str (default '')
            The description of the analytic
        mlflow_uri : str or None (default None)
            MLFlow URI to use, if applicable
        mlflow_user : str or None (default None)
            MLFlow user to use, if applicable
        mlflow_token : str or None (default None)
            MLFlow token to use, if applicable
        owner : str or None (default None)
            The owner of the model
        url : str (default '*')
            URL or URL pattern to match
        auto_run : bool (default False)
            Whether to automatically run this file when on a valid page
        documentation_link : str (default '')
            If provided, a URL to the model card for the .air file
        """
        super().__init__()
        self.name = name
        self.stage = stage
        self.version = version
        self.description = description
        self.mlflow_uri = mlflow_uri
        self.mlflow_user = mlflow_user
        self.mlflow_token = mlflow_token
        self.owner = owner
        self.url = url
        self.auto_run = auto_run
        self.documentation_link = documentation_link
        self.nodes = []

    # name
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = str(value)

    # stage
    @property
    def stage(self):
        return self._stage

    @stage.setter
    def stage(self, value):
        if value not in ALLOWED_STAGES:
            raise ValueError(
                f'stage must be one of {ALLOWED_STAGES}, got {value}')
        self._stage = value

    # version
    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, value):
        self._version = str(value) if value is not None else value

    # description
    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = str(value)

    # mlflow_uri
    @property
    def mlflow_uri(self):
        return self._mlflow_uri

    @mlflow_uri.setter
    def mlflow_uri(self, value):
        self._mlflow_uri = value

    # mlflow_user
    @property
    def mlflow_user(self):
        return self._mlflow_user

    @mlflow_user.setter
    def mlflow_user(self, value):
        self._mlflow_user = value

    # mlflow_token
    @property
    def mlflow_token(self):
        return self._mlflow_token

    @mlflow_token.setter
    def mlflow_token(self, value):
        self._mlflow_token = value

    # owner
    @property
    def owner(self):
        return self._owner

    @owner.setter
    def owner(self, value):
        self._owner = str(value) if value is not None else value

    # url
    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        if not isinstance(value, str):
            raise ValueError('url must be string')
        self._url = value

    # auto_run
    @property
    def auto_run(self):
        return self._auto_run

    @auto_run.setter
    def auto_run(self, value):
        if not isinstance(value, bool):
            raise TypeError('auto_run must be Boolean valued')
        self._auto_run = value

    # documentation_link
    @property
    def documentation_link(self):
        return self._documentation_link

    @documentation_link.setter
    def documentation_link(self, value):
        if not isinstance(value, str):
            raise TypeError('documentation_link must be str')
        self._documentation_link = value

    def add_node(self, step: BaseObject, dependencies: Union[int, list] = None) -> int:
        """
        Add a node to the configuration graph

        Parameters
        ----------
        step : aisquared configuration step
            The step to add
        dependencies : int, list of int, or None
            The ids of nodes which must be run before the
            added node

        Returns
        -------
        node_id : int
            The integer id of the node that is added
        """
        if not isinstance(step, BaseObject):
            raise TypeError(
                'Each node in the configuration graph should be an aisquared configuration step')
        if not (isinstance(dependencies, int) or dependencies is None):
            if not isinstance(dependencies, list) or not all([isinstance(dep, int) for dep in dependencies]):
                raise ValueError(
                    'dependencies must be integer or list of integers')

        id = len(self.nodes)
        self.nodes.append(
            {
                'id': id,
                'dependencies': dependencies,
                'step': step.to_dict()
            }
        )
        return id

    def get_filenames(self) -> list:
        """
        Get filenames for all models in the configuration
        """
        filenames = []
        for node in self.nodes:
            if node['step']['className'] in LOCAL_CLASSES:
                filenames.append(node['step']['params']['path'])
        return filenames

    def to_dict(self) -> dict:
        """
        Get the object as a dictionary
        """
        return {
            'className': 'GraphConfiguration',
            'params': {
                'name': self.name,
                'stage': self.stage,
                'version': self.version,
                'description': self.description,
                'mlflowUri': self.mlflow_uri,
                'mlflowUser': self.mlflow_user,
                'mlflowToken': self.mlflow_token,
                'owner': self.owner,
                'url': self.url,
                'autoRun': self.auto_run,
                'documentationLink': self.documentation_link
            },
            'nodes': self.nodes
        }

    def compile(self, filename: str = None, dtype: str = None) -> None:
        """
        Compile the object into a '.air' file, which can then be dragged and dropped into applications using the AI Squared JavaScript SDK

        Parameters
        ----------
        filename : path-like or None (default None)
            Filename to compile to. If None, defaults to '{NAME}.air', where {NAME} is the name of the analytic
        dtype : str or None (default None)
            The datatype to use for the model weights when using a Keras model. If None, defaults to 'float32'
        """
        if filename is None:
            filename = self.name + '.air'

        if dtype is None:
            dtype_map = None
        else:
            dtype_map = {dtype: '*'}

        dirname = os.path.join('.', os.path.splitext(filename)[0])

        # write the object as json config
        os.makedirs(dirname, exist_ok=True)
        with open(os.path.join(dirname, 'config.json'), 'w') as f:
            json.dump(self.to_dict(), f)

        # go through the files and copy them/make them tfjs files
        filenames = self.get_filenames()
        if len(filenames) != 0:
            for f in filenames:
                if os.path.splitext(f)[-1] in ['.h5', '.keras']:
                    model = tf.keras.models.load_model(f)
                    model_dir = os.path.join(dirname, os.path.split(f)[-1])
                    tfjs.converters.save_keras_model(
                        model, model_dir, quantization_dtype_map=dtype_map)
                else:
                    shutil.copy(f, dirname)

        # go through the entire directory of dirname, grab all files, and make
        # the archive file
        shutil.make_archive(filename, 'zip', dirname)

        # Move the archive file to just have .air
        shutil.move(filename + '.zip', filename)

        # Remove the temp directory
        shutil.rmtree(dirname, ignore_errors=True)



import pytest
import aisquared


def test_custom():
    obj = aisquared.config.CustomObject(
        'TestClass',
        test='foo',
        otherTest='bar'
    )
    assert obj.to_dict() == {
        'className': 'TestClass',
        'params': {
            'test': 'foo',
            'otherTest': 'bar'}
    }
    assert obj.path is None



from .AISquaredAPIException import AISquaredAPIException
from .additional_utils import _check_results_length
from aisquared.base import ENDPOINTS
import pandas as pd
import requests


def _list_model_users(
        url: str,
        headers: dict,
        model_id: str,
        as_df: bool
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    List users of a model

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    model_id : str
        The ID of the model
    as_df : bool
        Whether to return the data as a Pandas DataFrame
    """

    url = f'{url}/{ENDPOINTS["model"]}/{model_id}/users'

    with requests.Session() as sess:
        resp = sess.get(
            url,
            headers=headers
        )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())

    else:
        if as_df:
            df = pd.DataFrame(resp.json()['data'])
            _check_results_length(df)
            return df.sort_values(by='shared', ascending=False).reset_index(drop=True)

        return resp.json()


def _model_share_with_user(
        url: str,
        headers: dict,
        model_id: str,
        user_id: str,
        share: bool
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Share or unshare a model with a user

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    model_id : str
        The ID of the model
    user_id : str
        The ID for the user
    share : bool
        Whether to share (share = True) or unshare (share = False) a model with a user
    """

    url = f'{url}/{ENDPOINTS["model"]}/{model_id}/users/{user_id}'

    with requests.Session() as sess:
        if share:
            resp = sess.put(
                url,
                headers=headers
            )
        else:
            resp = sess.delete(
                url,
                headers=headers
            )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())
    return resp.ok


def _model_share_with_group(
        url: str,
        headers: dict,
        model_id: str,
        group_id: str,
        share: bool
):
    """
    NOT MEANT TO BE CALLED BY THE END USER

    Share or unshare a model with a group

    Parameters
    ----------
    url : string
        The base url to format
    headers : dict
        The headers used for authentication within the AI Squared platform
    model_id : str
        The ID of the model
    group_id : str
        The ID for the user
    share : bool
        Whether to share (share = True) or unshare (share = False) a model with a group
    """

    url = f'{url}/{ENDPOINTS["model"]}/{model_id}/groups/{group_id}'

    with requests.Session() as sess:
        if share:
            resp = sess.put(
                url,
                headers=headers
            )
        else:
            resp = sess.delete(
                url,
                headers=headers
            )

    if not resp.ok:
        raise AISquaredAPIException(resp.json())
    return resp.ok
